@@ -13,6 +13,7 @@ from typing import Any
 
 import snowflake.snowpark._internal.analyzer.analyzer_utils as analyzer_utils
 from snowflake.snowpark._internal.type_utils import convert_sp_to_sf_type
+from snowflake.snowpark._internal.utils import PythonObjJSONEncoder
 from snowflake.snowpark.types import (
     ArrayType,
     BinaryType,
@@ -21,13 +22,16 @@ from snowflake.snowpark.types import (
     DateType,
     DecimalType,
     GeographyType,
+    GeometryType,
     MapType,
     NullType,
     StringType,
     StructType,
+    TimestampTimeZone,
     TimestampType,
     TimeType,
     VariantType,
+    VectorType,
     _FractionalType,
     _IntegralType,
     _NumericType,
@@ -48,7 +52,7 @@ def to_sql(value: Any, datatype: DataType, from_values_statement: bool = False)
     # Handle null values
     if isinstance(
         datatype,
-        (NullType, ArrayType, MapType, StructType, GeographyType),
+        (NullType, ArrayType, MapType, StructType, GeographyType, GeometryType),
     ):
         if value is None:
             return "NULL"
@@ -63,10 +67,16 @@ def to_sql(value: Any, datatype: DataType, from_values_statement: bool = False)
             return "NULL :: FLOAT"
     if isinstance(datatype, StringType):
         if value is None:
-            return "NULL :: STRING"
+            return f"NULL :: {analyzer_utils.string(datatype.length)}"
     if isinstance(datatype, BooleanType):
         if value is None:
             return "NULL :: BOOLEAN"
+    if isinstance(datatype, VariantType):
+        if value is None:
+            return "NULL :: VARIANT"
+    if isinstance(datatype, VectorType):
+        if value is None:
+            return f"NULL :: VECTOR({datatype.element_type},{datatype.dimension})"
     if value is None:
         return "NULL"
 
@@ -76,7 +86,7 @@ def to_sql(value: Any, datatype: DataType, from_values_statement: bool = False)
         # the sql value has to be casted to make sure the varchar length
         # will not be limited.
         return (
-            f"{str_to_sql(value)} :: STRING"
+            f"{str_to_sql(value)} :: {analyzer_utils.string(datatype.length)}"
             if from_values_statement
             else str_to_sql(value)
         )
@@ -110,16 +120,20 @@ def to_sql(value: Any, datatype: DataType, from_values_statement: bool = False)
             return f"DATE '{value.isoformat()}'"
 
     if isinstance(datatype, TimestampType):
-        if isinstance(value, int):
-            # add value as microseconds to 1970-01-01 00:00:00.00.
-            target_time = datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(
-                microseconds=value
-            )
-            trimmed_ms = target_time.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
-            return f"TIMESTAMP '{trimmed_ms}'"
-        elif isinstance(value, datetime):
-            trimmed_ms = value.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
-            return f"TIMESTAMP '{trimmed_ms}'"
+        if isinstance(value, (int, datetime)):
+            if isinstance(value, int):
+                # add value as microseconds to 1970-01-01 00:00:00.00.
+                value = datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(
+                    microseconds=value
+                )
+            if datatype.tz == TimestampTimeZone.NTZ:
+                return f"'{value}'::TIMESTAMP_NTZ"
+            elif datatype.tz == TimestampTimeZone.LTZ:
+                return f"'{value}'::TIMESTAMP_LTZ"
+            elif datatype.tz == TimestampTimeZone.TZ:
+                return f"'{value}'::TIMESTAMP_TZ"
+            else:
+                return f"TIMESTAMP '{value}'"
 
     if isinstance(datatype, TimeType):
         if isinstance(value, time):
@@ -127,13 +141,20 @@ def to_sql(value: Any, datatype: DataType, from_values_statement: bool = False)
             return f"TIME('{trimmed_ms}')"
 
     if isinstance(value, (list, bytes, bytearray)) and isinstance(datatype, BinaryType):
-        return f"'{binascii.hexlify(value).decode()}' :: BINARY"
+        return f"'{binascii.hexlify(bytes(value)).decode()}' :: BINARY"
 
     if isinstance(value, (list, tuple, array)) and isinstance(datatype, ArrayType):
-        return f"PARSE_JSON({str_to_sql(json.dumps(value))}) :: ARRAY"
+        return f"PARSE_JSON({str_to_sql(json.dumps(value, cls=PythonObjJSONEncoder))}) :: ARRAY"
 
     if isinstance(value, dict) and isinstance(datatype, MapType):
-        return f"PARSE_JSON({str_to_sql(json.dumps(value))}) :: OBJECT"
+        return f"PARSE_JSON({str_to_sql(json.dumps(value, cls=PythonObjJSONEncoder))}) :: OBJECT"
+
+    if isinstance(datatype, VariantType):
+        # PARSE_JSON returns VARIANT, so no need to append :: VARIANT here explicitly.
+        return f"PARSE_JSON({str_to_sql(json.dumps(value, cls=PythonObjJSONEncoder))})"
+
+    if isinstance(datatype, VectorType):
+        return f"{value} :: VECTOR({datatype.element_type},{datatype.dimension})"
 
     raise TypeError(f"Unsupported datatype {datatype}, value {value} by to_sql()")
 
@@ -142,6 +163,8 @@ def schema_expression(data_type: DataType, is_nullable: bool) -> str:
     if is_nullable:
         if isinstance(data_type, GeographyType):
             return "TRY_TO_GEOGRAPHY(NULL)"
+        if isinstance(data_type, GeometryType):
+            return "TRY_TO_GEOMETRY(NULL)"
         if isinstance(data_type, ArrayType):
             return "PARSE_JSON('NULL') :: ARRAY"
         if isinstance(data_type, MapType):
@@ -153,9 +176,9 @@ def schema_expression(data_type: DataType, is_nullable: bool) -> str:
     if isinstance(data_type, _NumericType):
         return "0 :: " + convert_sp_to_sf_type(data_type)
     if isinstance(data_type, StringType):
-        return "'a' :: STRING"
+        return f"'a' :: {analyzer_utils.string(data_type.length)}"
     if isinstance(data_type, BinaryType):
-        return "to_binary(hex_encode(1))"
+        return "'01' :: BINARY"
     if isinstance(data_type, DateType):
         return "date('2020-9-16')"
     if isinstance(data_type, BooleanType):
@@ -163,7 +186,14 @@ def schema_expression(data_type: DataType, is_nullable: bool) -> str:
     if isinstance(data_type, TimeType):
         return "to_time('04:15:29.999')"
     if isinstance(data_type, TimestampType):
-        return "to_timestamp_ntz('2020-09-16 06:30:00')"
+        if data_type.tz == TimestampTimeZone.NTZ:
+            return "to_timestamp_ntz('2020-09-16 06:30:00')"
+        elif data_type.tz == TimestampTimeZone.LTZ:
+            return "to_timestamp_ltz('2020-09-16 06:30:00')"
+        elif data_type.tz == TimestampTimeZone.TZ:
+            return "to_timestamp_tz('2020-09-16 06:30:00')"
+        else:
+            return "to_timestamp('2020-09-16 06:30:00')"
     if isinstance(data_type, ArrayType):
         return "to_array(0)"
     if isinstance(data_type, MapType):
@@ -172,6 +202,17 @@ def schema_expression(data_type: DataType, is_nullable: bool) -> str:
         return "to_variant(0)"
     if isinstance(data_type, GeographyType):
         return "to_geography('POINT(-122.35 37.55)')"
+    if isinstance(data_type, GeometryType):
+        return "to_geometry('POINT(-122.35 37.55)')"
+    if isinstance(data_type, VectorType):
+        if data_type.element_type == "int":
+            zero = int(0)
+        elif data_type.element_type == "float":
+            zero = float(0)
+        else:
+            raise TypeError(f"Invalid vector element type: {data_type.element_type}")
+        values = [i + zero for i in range(data_type.dimension)]
+        return f"{values} :: VECTOR({data_type.element_type},{data_type.dimension})"
     raise Exception(f"Unsupported data type: {data_type.__class__.__name__}")
 
 
