@@ -6,15 +6,31 @@
 import re
 import sys
 import uuid
+from collections import defaultdict
 from functools import cached_property
-from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    DefaultDict,
+    Dict,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+)
 
-from snowflake.snowpark._internal.analyzer.table_function import GeneratorTableFunction
+from snowflake.snowpark._internal.analyzer.table_function import (
+    GeneratorTableFunction,
+    TableFunctionRelation,
+)
 
 if TYPE_CHECKING:
     from snowflake.snowpark._internal.analyzer.select_statement import (
         Selectable,
     )  # pragma: no cover
+    import snowflake.snowpark.session
+    import snowflake.snowpark.dataframe
 
 import snowflake.connector
 import snowflake.snowpark
@@ -25,6 +41,7 @@ from snowflake.snowpark._internal.analyzer.analyzer_utils import (
     copy_into_location,
     copy_into_table,
     create_file_format_statement,
+    create_or_replace_dynamic_table_statement,
     create_or_replace_view_statement,
     create_table_as_select_statement,
     create_table_statement,
@@ -41,6 +58,7 @@ from snowflake.snowpark._internal.analyzer.analyzer_utils import (
     merge_statement,
     pivot_statement,
     project_statement,
+    rename_statement,
     result_scan_statement,
     sample_statement,
     schema_cast_named,
@@ -78,9 +96,9 @@ from snowflake.snowpark.types import StructType
 # Python 3.8 needs to use typing.Iterable because collections.abc.Iterable is not subscriptable
 # Python 3.9 can use both
 # Python 3.10 needs to use collections.abc.Iterable because typing.Iterable is removed
-try:
+if sys.version_info <= (3, 9):
     from typing import Iterable
-except ImportError:
+else:
     from collections.abc import Iterable
 
 
@@ -97,10 +115,9 @@ class SnowflakePlan(LogicalPlan):
                 try:
                     return func(*args, **kwargs)
                 except snowflake.connector.errors.ProgrammingError as e:
-                    query = None
-                    if "query" in e.__dict__:
-                        query = e.__getattribute__("query")
+                    query = getattr(e, "query", None)
                     tb = sys.exc_info()[2]
+                    assert e.msg is not None
                     if "unexpected 'as'" in e.msg.lower():
                         ne = SnowparkClientExceptionMessages.SQL_PYTHON_REPORT_UNEXPECTED_ALIAS(
                             query
@@ -175,10 +192,14 @@ class SnowflakePlan(LogicalPlan):
         schema_query: str,
         post_actions: Optional[List["Query"]] = None,
         expr_to_alias: Optional[Dict[uuid.UUID, str]] = None,
-        session: Optional["snowflake.snowpark.session.Session"] = None,
         source_plan: Optional[LogicalPlan] = None,
         is_ddl_on_temp_object: bool = False,
         api_calls: Optional[List[Dict]] = None,
+        df_aliased_col_name_to_real_col_name: Optional[
+            DefaultDict[str, Dict[str, str]]
+        ] = None,
+        *,
+        session: "snowflake.snowpark.session.Session",
     ) -> None:
         super().__init__()
         self.queries = queries
@@ -191,6 +212,14 @@ class SnowflakePlan(LogicalPlan):
         # We need to copy this list since we don't want to change it for the
         # previous SnowflakePlan objects
         self.api_calls = api_calls.copy() if api_calls else []
+        self._output_dict = None
+        # Used for dataframe alias
+        if df_aliased_col_name_to_real_col_name:
+            self.df_aliased_col_name_to_real_col_name = (
+                df_aliased_col_name_to_real_col_name
+            )
+        else:
+            self.df_aliased_col_name_to_real_col_name = defaultdict(dict)
 
     def with_subqueries(self, subquery_plans: List["SnowflakePlan"]) -> "SnowflakePlan":
         pre_queries = self.queries[:-1]
@@ -218,6 +247,7 @@ class SnowflakePlan(LogicalPlan):
             session=self.session,
             source_plan=self.source_plan,
             api_calls=api_calls,
+            df_aliased_col_name_to_real_col_name=self.df_aliased_col_name_to_real_col_name,
         )
 
     @cached_property
@@ -230,16 +260,25 @@ class SnowflakePlan(LogicalPlan):
     def output(self) -> List[Attribute]:
         return [Attribute(a.name, a.datatype, a.nullable) for a in self.attributes]
 
+    @property
+    def output_dict(self) -> Dict[str, Any]:
+        if self._output_dict is None:
+            self._output_dict = {
+                attr.name: (attr.datatype, attr.nullable) for attr in self.output
+            }
+        return self._output_dict
+
     def __copy__(self) -> "SnowflakePlan":
         return SnowflakePlan(
             self.queries.copy() if self.queries else [],
             self.schema_query,
             self.post_actions.copy() if self.post_actions else None,
             dict(self.expr_to_alias) if self.expr_to_alias else None,
-            self.session,
             self.source_plan,
             self.is_ddl_on_temp_object,
             self.api_calls.copy() if self.api_calls else None,
+            self.df_aliased_col_name_to_real_col_name,
+            session=self.session,
         )
 
     def add_aliases(self, to_add: Dict) -> None:
@@ -265,6 +304,7 @@ class SnowflakePlanBuilder:
                 sql_generator(select_child.queries[-1].sql),
                 query_id_place_holder="",
                 is_ddl_on_temp_object=is_ddl_on_temp_object,
+                params=select_child.queries[-1].params,
             )
         ]
         new_schema_query = (
@@ -276,30 +316,35 @@ class SnowflakePlanBuilder:
             new_schema_query,
             select_child.post_actions,
             select_child.expr_to_alias,
-            self.session,
             source_plan,
             is_ddl_on_temp_object,
             api_calls=select_child.api_calls,
+            df_aliased_col_name_to_real_col_name=child.df_aliased_col_name_to_real_col_name,
+            session=self.session,
         )
 
     @SnowflakePlan.Decorator.wrap_exception
     def build_from_multiple_queries(
         self,
-        multi_sql_generator: Callable[[str], str],
+        multi_sql_generator: Callable[["Query"], List["Query"]],
         child: SnowflakePlan,
         source_plan: Optional[LogicalPlan],
-        schema_query: Optional[str] = None,
+        schema_query: str,
         is_ddl_on_temp_object: bool = False,
     ) -> SnowflakePlan:
         select_child = self.add_result_scan_if_not_select(child)
         queries = select_child.queries[0:-1] + [
-            Query(msg, is_ddl_on_temp_object=is_ddl_on_temp_object)
-            for msg in multi_sql_generator(select_child.queries[-1].sql)
+            Query(
+                query.sql,
+                is_ddl_on_temp_object=is_ddl_on_temp_object,
+                params=query.params,
+            )
+            for query in multi_sql_generator(select_child.queries[-1])
         ]
         new_schema_query = (
             schema_query
             if schema_query is not None
-            else multi_sql_generator(child.schema_query)[-1]
+            else multi_sql_generator(Query(child.schema_query))[-1].sql
         )
 
         return SnowflakePlan(
@@ -307,9 +352,9 @@ class SnowflakePlanBuilder:
             new_schema_query,
             select_child.post_actions,
             select_child.expr_to_alias,
-            self.session,
             source_plan,
             api_calls=select_child.api_calls,
+            session=self.session,
         )
 
     @SnowflakePlan.Decorator.wrap_exception
@@ -330,7 +375,10 @@ class SnowflakePlanBuilder:
                     sql_generator(
                         select_left.queries[-1].sql, select_right.queries[-1].sql
                     ),
-                    None,
+                    params=[
+                        *select_left.queries[-1].params,
+                        *select_right.queries[-1].params,
+                    ],
                 )
             ]
         )
@@ -357,9 +405,9 @@ class SnowflakePlanBuilder:
             schema_query,
             select_left.post_actions + select_right.post_actions,
             new_expr_to_alias,
-            self.session,
             source_plan,
             api_calls=api_calls,
+            session=self.session,
         )
 
     def query(
@@ -367,10 +415,12 @@ class SnowflakePlanBuilder:
         sql: str,
         source_plan: Optional[LogicalPlan],
         api_calls: Optional[List[Dict]] = None,
+        params: Optional[Sequence[Any]] = None,
+        schema_query: Optional[str] = None,
     ) -> SnowflakePlan:
         return SnowflakePlan(
-            queries=[Query(sql)],
-            schema_query=sql,
+            queries=[Query(sql, params=params)],
+            schema_query=schema_query or sql,
             session=self.session,
             source_plan=source_plan,
             api_calls=api_calls,
@@ -510,17 +560,67 @@ class SnowflakePlanBuilder:
 
     def save_as_table(
         self,
-        table_name: str,
+        table_name: Iterable[str],
         column_names: Optional[Iterable[str]],
         mode: SaveMode,
         table_type: str,
+        clustering_keys: Iterable[str],
         child: SnowflakePlan,
     ) -> SnowflakePlan:
+        full_table_name = ".".join(table_name)
+
+        # here get the column definition from the child attributes. In certain cases we have
+        # the attributes set to ($1, VariantType()) which cannot be used as valid column name
+        # in save as table. So we rename ${number} with COL{number}.
+        hidden_column_pattern = r"\$(\d+)"
+        column_definition_with_hidden_columns = attribute_to_schema_string(
+            child.attributes
+        )
+        column_definition = re.sub(
+            hidden_column_pattern,
+            lambda match: f"COL{match.group(1)}",
+            column_definition_with_hidden_columns,
+        )
+
+        def get_create_and_insert_plan(child: SnowflakePlan, replace=False, error=True):
+            create_table = create_table_statement(
+                full_table_name,
+                column_definition,
+                replace=replace,
+                error=error,
+                table_type=table_type,
+                clustering_key=clustering_keys,
+            )
+
+            # so that dataframes created from non-select statements,
+            # such as table sprocs, work
+            child = self.add_result_scan_if_not_select(child)
+            return SnowflakePlan(
+                [
+                    *child.queries[0:-1],
+                    Query(create_table),
+                    Query(
+                        insert_into_statement(
+                            table_name=full_table_name,
+                            child=child.queries[-1].sql,
+                            column_names=column_names,
+                        ),
+                        params=child.queries[-1].params,
+                    ),
+                ],
+                create_table,
+                child.post_actions,
+                {},
+                None,
+                api_calls=child.api_calls,
+                session=self.session,
+            )
+
         if mode == SaveMode.APPEND:
             if self.session._table_exists(table_name):
                 return self.build(
                     lambda x: insert_into_statement(
-                        table_name=table_name,
+                        table_name=full_table_name,
                         child=x,
                         column_names=column_names,
                     ),
@@ -528,36 +628,16 @@ class SnowflakePlanBuilder:
                     None,
                 )
             else:
-                create_table = create_table_statement(
-                    table_name,
-                    attribute_to_schema_string(child.attributes),
-                    error=False,
-                    table_type=table_type,
-                )
-
-                return SnowflakePlan(
-                    [
-                        *child.queries[0:-1],
-                        Query(create_table),
-                        Query(
-                            insert_into_statement(
-                                table_name=table_name,
-                                child=child.queries[-1].sql,
-                                column_names=column_names,
-                            )
-                        ),
-                    ],
-                    create_table,
-                    child.post_actions,
-                    {},
-                    self.session,
-                    None,
-                    api_calls=child.api_calls,
-                )
+                return get_create_and_insert_plan(child, replace=False, error=False)
         elif mode == SaveMode.OVERWRITE:
             return self.build(
                 lambda x: create_table_as_select_statement(
-                    table_name, x, replace=True, table_type=table_type
+                    full_table_name,
+                    x,
+                    column_definition,
+                    replace=True,
+                    table_type=table_type,
+                    clustering_key=clustering_keys,
                 ),
                 child,
                 None,
@@ -565,7 +645,12 @@ class SnowflakePlanBuilder:
         elif mode == SaveMode.IGNORE:
             return self.build(
                 lambda x: create_table_as_select_statement(
-                    table_name, x, error=False, table_type=table_type
+                    full_table_name,
+                    x,
+                    column_definition,
+                    error=False,
+                    table_type=table_type,
+                    clustering_key=clustering_keys,
                 ),
                 child,
                 None,
@@ -573,7 +658,11 @@ class SnowflakePlanBuilder:
         elif mode == SaveMode.ERROR_IF_EXISTS:
             return self.build(
                 lambda x: create_table_as_select_statement(
-                    table_name, x, table_type=table_type
+                    full_table_name,
+                    x,
+                    column_definition,
+                    table_type=table_type,
+                    clustering_key=clustering_keys,
                 ),
                 child,
                 None,
@@ -621,6 +710,18 @@ class SnowflakePlanBuilder:
             source_plan,
         )
 
+    def rename(
+        self,
+        column_map: Dict[str, str],
+        child: SnowflakePlan,
+        source_plan: Optional[LogicalPlan],
+    ) -> SnowflakePlan:
+        return self.build(
+            lambda x: rename_statement(column_map, x),
+            child,
+            source_plan,
+        )
+
     def create_or_replace_view(
         self, name: str, child: SnowflakePlan, is_temp: bool
     ) -> SnowflakePlan:
@@ -636,6 +737,27 @@ class SnowflakePlanBuilder:
             None,
         )
 
+    def create_or_replace_dynamic_table(
+        self,
+        name: str,
+        warehouse: str,
+        lag: str,
+        child: SnowflakePlan,
+    ) -> SnowflakePlan:
+        if len(child.queries) != 1:
+            raise SnowparkClientExceptionMessages.PLAN_CREATE_DYNAMIC_TABLE_FROM_DDL_DML_OPERATIONS()
+
+        if not is_sql_select_statement(child.queries[0].sql.lower().strip()):
+            raise SnowparkClientExceptionMessages.PLAN_CREATE_DYNAMIC_TABLE_FROM_SELECT_ONLY()
+
+        return self.build(
+            lambda x: create_or_replace_dynamic_table_statement(
+                name, warehouse, lag, x
+            ),
+            child,
+            None,
+        )
+
     def create_temp_table(
         self,
         name: str,
@@ -664,11 +786,11 @@ class SnowflakePlanBuilder:
         session,
         name: str,
         schema_query: str,
-        query: str,
+        query: "Query",
         *,
         use_scoped_temp_objects: bool = False,
         is_generated: bool = False,
-    ) -> List[str]:
+    ) -> List["Query"]:
         attributes = session._get_result_attributes(schema_query)
         create_table = create_table_statement(
             name,
@@ -679,8 +801,13 @@ class SnowflakePlanBuilder:
         )
 
         return [
-            create_table,
-            insert_into_statement(table_name=name, column_names=None, child=query),
+            Query(create_table),
+            Query(
+                insert_into_statement(
+                    table_name=name, column_names=None, child=query.sql
+                ),
+                params=query.params,
+            ),
         ]
 
     def read_file(
@@ -692,10 +819,13 @@ class SnowflakePlanBuilder:
         schema: List[Attribute],
         schema_to_cast: Optional[List[Tuple[str, str]]] = None,
         transformations: Optional[List[str]] = None,
+        metadata_project: Optional[List[str]] = None,
+        metadata_schema: Optional[List[Attribute]] = None,
     ):
         format_type_options, copy_options = get_copy_into_table_options(options)
-        pattern = options.get("PATTERN", None)
+        pattern = options.get("PATTERN")
         # Can only infer the schema for parquet, orc and avro
+        # csv and json in preview
         infer_schema = (
             options.get("INFER_SCHEMA", True)
             if format in INFER_SCHEMA_FORMAT_TYPES
@@ -705,6 +835,16 @@ class SnowflakePlanBuilder:
         if pattern:
             self.session._conn._telemetry_client.send_copy_pattern_telemetry()
 
+        if format_type_options.get("PARSE_HEADER", False):
+            # This option is only available for CSV file format
+            # The options is used when specified with INFER_SCHEMA( ..., FILE_FORMAT => (.., PARSE_HEADER)) see
+            # https://docs.snowflake.com/en/sql-reference/sql/create-file-format#format-type-options-formattypeoptions
+            # PARSE_HEADER does not work with FILE_FORMAT when used with SELECT FROM LOCATION(FILE_FORMAT ...). Thus,
+            # if user has set option("PARSE_HEADER", True), we have already read the header in
+            # DataframeReader._infer_schema_for_file_format so now we must set skip_header = 1 to skip the header line.
+            format_type_options["SKIP_HEADER"] = 1
+        format_type_options.pop("PARSE_HEADER", None)
+
         if not copy_options:  # use select
             queries: List[Query] = []
             post_queries: List[Query] = []
@@ -738,25 +878,30 @@ class SnowflakePlanBuilder:
             else:
                 format_name = options["FORMAT_NAME"]
 
+            if infer_schema:
+                assert schema_to_cast is not None
+                schema_project: List[str] = schema_cast_named(schema_to_cast)
+            else:
+                schema_project: List[str] = schema_cast_seq(schema)
+
             queries.append(
                 Query(
                     select_from_path_with_format_statement(
-                        schema_cast_named(schema_to_cast)
-                        if infer_schema
-                        else schema_cast_seq(schema),
+                        (metadata_project or []) + schema_project,
                         path,
                         format_name,
                         pattern,
                     )
                 )
             )
+
             return SnowflakePlan(
                 queries,
-                schema_value_statement(schema),
+                schema_value_statement((metadata_schema or []) + schema),
                 post_queries,
                 {},
-                self.session,
                 None,
+                session=self.session,
             )
         else:  # otherwise use COPY
             if "FORCE" in copy_options and str(copy_options["FORCE"]).lower() != "true":
@@ -830,31 +975,33 @@ class SnowflakePlanBuilder:
                 schema_value_statement(schema),
                 post_actions,
                 {},
-                self.session,
                 None,
+                session=self.session,
             )
 
     def copy_into_table(
         self,
         file_format: str,
-        table_name: str,
-        path: Optional[str] = None,
+        table_name: Iterable[str],
+        path: str,
         files: Optional[str] = None,
         pattern: Optional[str] = None,
-        format_type_options: Optional[Dict[str, Any]] = None,
-        copy_options: Optional[Dict[str, Any]] = None,
         validation_mode: Optional[str] = None,
         column_names: Optional[List[str]] = None,
         transformations: Optional[List[str]] = None,
         user_schema: Optional[StructType] = None,
         create_table_from_infer_schema: bool = False,
+        *,
+        copy_options: Dict[str, Any],
+        format_type_options: Dict[str, Any],
     ) -> SnowflakePlan:
         # tracking usage of pattern, will refactor this function in future
         if pattern:
             self.session._conn._telemetry_client.send_copy_pattern_telemetry()
 
+        full_table_name = ".".join(table_name)
         copy_command = copy_into_table(
-            table_name=table_name,
+            table_name=full_table_name,
             file_path=path,
             files=files,
             file_format_type=file_format,
@@ -878,7 +1025,7 @@ class SnowflakePlanBuilder:
             queries = [
                 Query(
                     create_table_statement(
-                        table_name,
+                        full_table_name,
                         attribute_to_schema_string(attributes),
                     ),
                     # This is an exception. The principle is to avoid surprising behavior and most of the time
@@ -890,9 +1037,9 @@ class SnowflakePlanBuilder:
             ]
         else:
             raise SnowparkClientExceptionMessages.DF_COPY_INTO_CANNOT_CREATE_TABLE(
-                table_name
+                full_table_name
             )
-        return SnowflakePlan(queries, copy_command, [], {}, self.session, None)
+        return SnowflakePlan(queries, copy_command, [], {}, None, session=self.session)
 
     def copy_into_location(
         self,
@@ -1005,7 +1152,7 @@ class SnowflakePlanBuilder:
         )
 
     def from_table_function(
-        self, func: str, source_plan: Optional[LogicalPlan]
+        self, func: str, source_plan: TableFunctionRelation
     ) -> SnowflakePlan:
         if isinstance(source_plan.table_function, GeneratorTableFunction):
             return self.query(
@@ -1015,10 +1162,18 @@ class SnowflakePlanBuilder:
         return self.query(table_function_statement(func), None)
 
     def join_table_function(
-        self, func: str, child: SnowflakePlan, source_plan: Optional[LogicalPlan]
+        self,
+        func: str,
+        child: SnowflakePlan,
+        source_plan: Optional[LogicalPlan],
+        left_cols: List[str],
+        right_cols: List[str],
+        use_constant_subquery_alias: bool,
     ) -> SnowflakePlan:
         return self.build(
-            lambda x: join_table_function_statement(func, x),
+            lambda x: join_table_function_statement(
+                func, x, left_cols, right_cols, use_constant_subquery_alias
+            ),
             child,
             source_plan,
         )
@@ -1035,7 +1190,6 @@ class SnowflakePlanBuilder:
             new_queries = plan.queries + [
                 Query(
                     result_scan_statement(plan.queries[-1].query_id_place_holder),
-                    None,
                 )
             ]
             return SnowflakePlan(
@@ -1043,9 +1197,9 @@ class SnowflakePlanBuilder:
                 schema_value_statement(plan.attributes),
                 plan.post_actions,
                 plan.expr_to_alias,
-                self.session,
                 plan.source_plan,
                 api_calls=plan.api_calls,
+                session=self.session,
             )
 
 
@@ -1053,8 +1207,10 @@ class Query:
     def __init__(
         self,
         sql: str,
+        *,
         query_id_place_holder: Optional[str] = None,
         is_ddl_on_temp_object: bool = False,
+        params: Optional[Sequence[Any]] = None,
     ) -> None:
         self.sql = sql
         self.query_id_place_holder = (
@@ -1063,9 +1219,17 @@ class Query:
             else f"query_id_place_holder_{generate_random_alphanumeric()}"
         )
         self.is_ddl_on_temp_object = is_ddl_on_temp_object
+        self.params = params or []
 
     def __repr__(self) -> str:
-        return f"Query({self.sql!r}, {self.query_id_place_holder!r}, {self.is_ddl_on_temp_object})"
+        return (
+            "Query("
+            + f"{self.sql!r}, "
+            + f"query_id_place_holder={self.query_id_place_holder!r}, "
+            + f"is_ddl_on_temp_object={self.is_ddl_on_temp_object}, "
+            + f"params={self.params}"
+            + ")"
+        )
 
     def __eq__(self, other: "Query") -> bool:
         return (
