@@ -2,7 +2,7 @@
 # Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
 #
 
-from typing import Optional, Set
+from typing import AbstractSet, Optional, Type
 
 from snowflake.snowpark._internal.analyzer.expression import (
     Expression,
@@ -24,7 +24,7 @@ class NullsLast(NullOrdering):
 
 class SortDirection:
     sql: str
-    default_null_ordering: NullOrdering
+    default_null_ordering: Type[NullOrdering]
 
 
 class Ascending(SortDirection):
@@ -45,6 +45,7 @@ class SortOrder(Expression):
         null_ordering: Optional[NullOrdering] = None,
     ) -> None:
         super().__init__(child)
+        self.child: Expression
         self.direction = direction
         self.null_ordering = (
             null_ordering if null_ordering else direction.default_null_ordering
@@ -52,5 +53,5 @@ class SortOrder(Expression):
         self.datatype = child.datatype
         self.nullable = child.nullable
 
-    def dependent_column_names(self) -> Optional[Set[str]]:
+    def dependent_column_names(self) -> Optional[AbstractSet[str]]:
         return derive_dependent_columns(self.child)
