@@ -1,13 +1,22 @@
 #
-# Copyright (c) 2012-2022 Snowflake Computing Inc. All rights reserved.
+# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
 #
 
-from typing import Dict, Iterable, List, Optional
+import sys
+from typing import Dict, List, Optional
 
 from snowflake.snowpark._internal.analyzer.expression import Expression
 from snowflake.snowpark._internal.analyzer.snowflake_plan_node import LogicalPlan
 from snowflake.snowpark._internal.analyzer.sort_expression import SortOrder
 
+# Python 3.8 needs to use typing.Iterable because collections.abc.Iterable is not subscriptable
+# Python 3.9 can use both
+# Python 3.10 needs to use collections.abc.Iterable because typing.Iterable is removed
+if sys.version_info <= (3, 9):
+    from typing import Iterable
+else:
+    from collections.abc import Iterable
+
 
 class TableFunctionPartitionSpecDefinition(Expression):
     def __init__(
@@ -86,11 +95,17 @@ class TableFunctionRelation(LogicalPlan):
 
 class TableFunctionJoin(LogicalPlan):
     def __init__(
-        self, child: LogicalPlan, table_function: TableFunctionExpression
+        self,
+        child: LogicalPlan,
+        table_function: TableFunctionExpression,
+        left_cols: Optional[List[str]] = None,
+        right_cols: Optional[List[str]] = None,
     ) -> None:
         super().__init__()
         self.children = [child]
         self.table_function = table_function
+        self.left_cols = left_cols if left_cols is not None else ["*"]
+        self.right_cols = right_cols if right_cols is not None else ["*"]
 
 
 class Lateral(LogicalPlan):
