@@ -1,10 +1,14 @@
 #
-# Copyright (c) 2012-2022 Snowflake Computing Inc. All rights reserved.
+# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
 #
+
 from typing import Dict, List, Optional
 
 from snowflake.snowpark._internal.analyzer.expression import Expression
-from snowflake.snowpark._internal.analyzer.snowflake_plan import LogicalPlan
+from snowflake.snowpark._internal.analyzer.snowflake_plan import (
+    LogicalPlan,
+    SnowflakePlan,
+)
 
 
 class MergeExpression(Expression):
@@ -43,7 +47,7 @@ class TableUpdate(LogicalPlan):
         table_name: str,
         assignments: Dict[Expression, Expression],
         condition: Optional[Expression],
-        source_data: Optional[LogicalPlan],
+        source_data: Optional[SnowflakePlan],
     ) -> None:
         super().__init__()
         self.table_name = table_name
@@ -58,7 +62,7 @@ class TableDelete(LogicalPlan):
         self,
         table_name: str,
         condition: Optional[Expression],
-        source_data: Optional[LogicalPlan],
+        source_data: Optional[SnowflakePlan],
     ) -> None:
         super().__init__()
         self.table_name = table_name
@@ -71,7 +75,7 @@ class TableMerge(LogicalPlan):
     def __init__(
         self,
         table_name: str,
-        source: LogicalPlan,
+        source: SnowflakePlan,
         join_expr: Expression,
         clauses: List[Expression],
     ) -> None:
