@@ -4,11 +4,25 @@
 #
 
 import functools
+import inspect
 import os
 import sys
 import time
+import warnings
 from logging import getLogger
-from typing import IO, Any, Dict, Iterator, List, Optional, Set, Tuple, Union
+from typing import (
+    IO,
+    TYPE_CHECKING,
+    Any,
+    Dict,
+    Iterator,
+    List,
+    Optional,
+    Sequence,
+    Set,
+    Tuple,
+    Union,
+)
 
 from snowflake.connector import SnowflakeConnection, connect
 from snowflake.connector.constants import ENV_VAR_PARTNER, FIELD_ID_TO_NAME
@@ -17,13 +31,14 @@ from snowflake.connector.errors import NotSupportedError, ProgrammingError
 from snowflake.connector.network import ReauthenticationRequest
 from snowflake.connector.options import pandas
 from snowflake.snowpark._internal.analyzer.analyzer_utils import (
-    escape_quotes,
     quote_name_without_upper_casing,
 )
 from snowflake.snowpark._internal.analyzer.datatype_mapper import str_to_sql
 from snowflake.snowpark._internal.analyzer.expression import Attribute
 from snowflake.snowpark._internal.analyzer.schema_utils import (
     convert_result_meta_to_attribute,
+    get_new_description,
+    run_new_describe,
 )
 from snowflake.snowpark._internal.analyzer.snowflake_plan import (
     BatchInsertQuery,
@@ -32,6 +47,7 @@ from snowflake.snowpark._internal.analyzer.snowflake_plan import (
 from snowflake.snowpark._internal.error_message import SnowparkClientExceptionMessages
 from snowflake.snowpark._internal.telemetry import TelemetryClient
 from snowflake.snowpark._internal.utils import (
+    escape_quotes,
     get_application_name,
     get_version,
     is_in_stored_procedure,
@@ -45,6 +61,12 @@ from snowflake.snowpark.async_job import AsyncJob, _AsyncResultType
 from snowflake.snowpark.query_history import QueryHistory, QueryRecord
 from snowflake.snowpark.row import Row
 
+if TYPE_CHECKING:
+    try:
+        from snowflake.connector.cursor import ResultMetadataV2
+    except ImportError:
+        ResultMetadataV2 = ResultMetadata
+
 logger = getLogger(__name__)
 
 # parameters needed for usage tracking
@@ -142,6 +164,12 @@ class ServerConnection:
         # Snowpark session
         self._telemetry_client.send_session_created_telemetry(not bool(conn))
 
+        # check if cursor.execute supports _skip_upload_on_content_match
+        signature = inspect.signature(self._cursor.execute)
+        self._supports_skip_upload_on_content_match = (
+            "_skip_upload_on_content_match" in signature.parameters
+        )
+
     def _add_application_name(self) -> None:
         if PARAM_APPLICATION not in self._lower_case_parameters:
             # Mirrored from snowflake-connector-python/src/snowflake/connector/connection.py#L295
@@ -196,7 +224,7 @@ class ServerConnection:
 
     @SnowflakePlan.Decorator.wrap_exception
     def get_result_attributes(self, query: str) -> List[Attribute]:
-        return convert_result_meta_to_attribute(self._cursor.describe(query))
+        return convert_result_meta_to_attribute(run_new_describe(self._cursor, query))
 
     @_Decorator.log_msg_and_perf_telemetry("Uploading file to stage")
     def upload_file(
@@ -208,6 +236,7 @@ class ServerConnection:
         compress_data: bool = True,
         source_compression: str = "AUTO_DETECT",
         overwrite: bool = False,
+        skip_upload_on_content_match: bool = False,
     ) -> Optional[Dict[str, Any]]:
         if is_in_stored_procedure():  # pragma: no cover
             file_name = os.path.basename(path)
@@ -225,6 +254,10 @@ class ServerConnection:
                 raise ne.with_traceback(tb) from None
         else:
             uri = normalize_local_file(path)
+            if self._supports_skip_upload_on_content_match:
+                kwargs = {"_skip_upload_on_content_match": skip_upload_on_content_match}
+            else:
+                kwargs = {}
             return self.run_query(
                 _build_put_statement(
                     uri,
@@ -234,7 +267,8 @@ class ServerConnection:
                     compress_data,
                     source_compression,
                     overwrite,
-                )
+                ),
+                **kwargs,
             )
 
     @_Decorator.log_msg_and_perf_telemetry("Uploading stream to stage")
@@ -249,6 +283,7 @@ class ServerConnection:
         source_compression: str = "AUTO_DETECT",
         overwrite: bool = False,
         is_in_udf: bool = False,
+        skip_upload_on_content_match: bool = False,
     ) -> Optional[Dict[str, Any]]:
         uri = normalize_local_file(f"/tmp/placeholder/{dest_filename}")
         try:
@@ -267,6 +302,13 @@ class ServerConnection:
                     )
                     raise ne.with_traceback(tb) from None
             else:
+                if self._supports_skip_upload_on_content_match:
+                    kwargs = {
+                        "_skip_upload_on_content_match": skip_upload_on_content_match,
+                        "file_stream": input_stream,
+                    }
+                else:
+                    kwargs = {"file_stream": input_stream}
                 return self.run_query(
                     _build_put_statement(
                         uri,
@@ -277,7 +319,7 @@ class ServerConnection:
                         source_compression,
                         overwrite,
                     ),
-                    file_stream=input_stream,
+                    **kwargs,
                 )
         # If ValueError is raised and the stream is closed, we throw the error.
         # https://docs.python.org/3/library/io.html#io.IOBase.close
@@ -312,6 +354,8 @@ class ServerConnection:
         ] = None,  # this argument is currently only used by AsyncJob
         log_on_exception: bool = False,
         case_sensitive: bool = True,
+        params: Optional[Sequence[Any]] = None,
+        num_statements: Optional[int] = None,
         **kwargs,
     ) -> Union[Dict[str, Any], AsyncJob]:
         try:
@@ -321,13 +365,15 @@ class ServerConnection:
                     kwargs["_statement_params"] = {}
                 kwargs["_statement_params"]["SNOWPARK_SKIP_TXN_COMMIT_IN_DDL"] = True
             if block:
-                results_cursor = self._cursor.execute(query, **kwargs)
+                results_cursor = self._cursor.execute(query, params=params, **kwargs)
                 self.notify_query_listeners(
                     QueryRecord(results_cursor.sfqid, results_cursor.query)
                 )
                 logger.debug(f"Execute query [queryID: {results_cursor.sfqid}] {query}")
             else:
-                results_cursor = self._cursor.execute_async(query, **kwargs)
+                results_cursor = self._cursor.execute_async(
+                    query, params=params, num_statements=num_statements, **kwargs
+                )
                 self.notify_query_listeners(
                     QueryRecord(results_cursor["queryId"], query)
                 )
@@ -358,6 +404,7 @@ class ServerConnection:
                 async_job_plan.post_actions,
                 log_on_exception,
                 case_sensitive=case_sensitive,
+                num_statements=num_statements,
                 **kwargs,
             )
 
@@ -366,6 +413,7 @@ class ServerConnection:
         results_cursor: SnowflakeCursor,
         to_pandas: bool = False,
         to_iter: bool = False,
+        num_statements: Optional[int] = None,
     ) -> Dict[str, Any]:
         if to_pandas:
             try:
@@ -411,7 +459,13 @@ class ServerConnection:
     ) -> Union[
         List[Row], "pandas.DataFrame", Iterator[Row], Iterator["pandas.DataFrame"]
     ]:
-        if is_in_stored_procedure() and not block:  # pragma: no cover
+        if (
+            is_in_stored_procedure()
+            and not block
+            and not self._get_client_side_session_parameter(
+                "ENABLE_ASYNC_QUERY_IN_PYTHON_STORED_PROCS", False
+            )
+        ):  # pragma: no cover
             raise NotImplementedError(
                 "Async query is not supported in stored procedure yet"
             )
@@ -461,7 +515,7 @@ class ServerConnection:
                 str,
             ],
         ],
-        List[ResultMetadata],
+        Union[List[ResultMetadata], List["ResultMetadataV2"]],
     ]:
         action_id = plan.session._generate_new_action_id()
 
@@ -475,24 +529,20 @@ class ServerConnection:
                     break
             # since batch insert does not support async execution (? in the query), we handle it separately here
             if len(plan.queries) > 1 and not block and not is_batch_insert:
-                final_query = f"""EXECUTE IMMEDIATE $$
-DECLARE
-    res resultset;
-BEGIN
-    {";".join(q.sql for q in plan.queries[:-1])};
-    res := ({plan.queries[-1].sql});
-    return table(res);
-END;
-$$"""
-                # In multiple queries scenario, we are unable to get the query id of former query, so we replace
-                # place holder with fucntion last_query_id() here
+                params = []
+                final_queries = []
+                last_place_holder = None
                 for q in plan.queries:
-                    final_query = final_query.replace(
-                        f"'{q.query_id_place_holder}'", "LAST_QUERY_ID()"
+                    final_queries.append(
+                        q.sql.replace(f"'{last_place_holder}'", "LAST_QUERY_ID()")
+                        if last_place_holder
+                        else q.sql
                     )
+                    last_place_holder = q.query_id_place_holder
+                    params.extend(q.params)
 
                 result = self.run_query(
-                    final_query,
+                    ";".join(final_queries),
                     to_pandas,
                     to_iter,
                     is_ddl_on_temp_object=plan.queries[0].is_ddl_on_temp_object,
@@ -501,6 +551,8 @@ $$"""
                     async_job_plan=plan,
                     log_on_exception=log_on_exception,
                     case_sensitive=case_sensitive,
+                    num_statements=len(plan.queries),
+                    params=params,
                     **kwargs,
                 )
 
@@ -528,12 +580,13 @@ $$"""
                             async_job_plan=plan,
                             log_on_exception=log_on_exception,
                             case_sensitive=case_sensitive,
+                            params=query.params,
                             **kwargs,
                         )
                         placeholders[query.query_id_place_holder] = (
                             result["sfqid"] if not is_last else result.query_id
                         )
-                        result_meta = self._cursor.description
+                        result_meta = get_new_description(self._cursor)
                     if action_id < plan.session._last_canceled_id:
                         raise SnowparkClientExceptionMessages.SERVER_QUERY_IS_CANCELLED()
         finally:
@@ -559,8 +612,8 @@ $$"""
     ) -> Tuple[List[Row], List[Attribute]]:
         result_set, result_meta = self.get_result_set(plan, **kwargs)
         result = result_set_to_rows(result_set["data"])
-        meta = convert_result_meta_to_attribute(result_meta)
-        return result, meta
+        attributes = convert_result_meta_to_attribute(result_meta)
+        return result, attributes
 
     def get_result_query_id(self, plan: SnowflakePlan, **kwargs) -> str:
         # get the iterator such that the data is not fetched
@@ -600,10 +653,29 @@ $$"""
             )
         logger.debug("Execute batch insertion query %s", query)
 
+    def _get_client_side_session_parameter(self, name: str, default_value: Any) -> Any:
+        """It doesn't go to Snowflake to retrieve the session parameter.
+        Use this only when you know the Snowflake session parameter is sent to the client when a session/connection is created.
+        """
+        return (
+            self._conn._session_parameters.get(name, default_value)
+            if self._conn._session_parameters
+            else default_value
+        )
+
 
 def _fix_pandas_df_integer(
     pd_df: "pandas.DataFrame", results_cursor: SnowflakeCursor
 ) -> "pandas.DataFrame":
+    """The compiler does not make any guarantees about the return types - only that they will be large enough for the result.
+    As a result, the ResultMetadata may contain precision=38, scale=0 for result of a column which may only contain single
+    digit numbers. Then the returned pandas DataFrame has dtype "object" with a str value for that column instead of int64.
+
+    Based on the Result Metadata characteristics, this functions tries to make a best effort conversion to int64 without losing
+    precision.
+
+    We need to get rid of this workaround because this causes a performance hit.
+    """
     for column_metadata, pandas_dtype, pandas_col_name in zip(
         results_cursor.description, pd_df.dtypes, pd_df.columns
     ):
@@ -613,7 +685,25 @@ def _fix_pandas_df_integer(
             and column_metadata.scale == 0
             and not str(pandas_dtype).startswith("int")
         ):
-            pd_df[pandas_col_name] = pandas.to_numeric(
-                pd_df[pandas_col_name], downcast="integer"
-            )
+            # pandas.to_numeric raises "RuntimeWarning: invalid value encountered in cast"
+            # when it tried to downcast and loses precision. In this case, we try to convert to
+            # int64 using astype() and fallback to to_numeric if we were unsuccessful.
+            with warnings.catch_warnings(record=True) as warning_list:
+                pd_col_with_numeric_downcast = pandas.to_numeric(
+                    pd_df[pandas_col_name], downcast="integer"
+                )
+
+            if (
+                len(warning_list) == 1
+                and isinstance(warning_list[0].message, RuntimeWarning)
+                and warning_list[0].message.args
+                == ("invalid value encountered in cast",)
+            ):
+                try:
+                    pd_df[pandas_col_name] = pd_df[pandas_col_name].astype("int64")
+                except OverflowError:
+                    pd_df[pandas_col_name] = pd_col_with_numeric_downcast
+            else:
+                pd_df[pandas_col_name] = pd_col_with_numeric_downcast
+
     return pd_df
