@@ -1,7 +1,7 @@
 #
 # Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
 #
-
+import collections.abc
 import io
 import os
 import pickle
@@ -25,12 +25,15 @@ from typing import (
 import cloudpickle
 
 import snowflake.snowpark
-from snowflake.snowpark._internal import code_generation
+from snowflake.connector.options import installed_pandas, pandas
+from snowflake.snowpark._internal import code_generation, type_utils
 from snowflake.snowpark._internal.analyzer.datatype_mapper import to_sql
 from snowflake.snowpark._internal.telemetry import TelemetryField
 from snowflake.snowpark._internal.type_utils import (
+    NoneType,
     convert_sp_to_sf_type,
     infer_type,
+    python_type_str_to_object,
     python_type_to_snow_type,
     retrieve_func_type_hints_from_source,
 )
@@ -45,19 +48,21 @@ from snowflake.snowpark._internal.utils import (
     unwrap_stage_location_single_quote,
     validate_object_name,
 )
-from snowflake.snowpark.types import (
-    DataType,
-    PandasDataFrameType,
-    PandasSeriesType,
-    StructType,
-)
+from snowflake.snowpark.types import DataType, StructField, StructType
+
+if installed_pandas:
+    from snowflake.snowpark.types import (
+        PandasDataFrame,
+        PandasDataFrameType,
+        PandasSeriesType,
+    )
 
 # Python 3.8 needs to use typing.Iterable because collections.abc.Iterable is not subscriptable
 # Python 3.9 can use both
 # Python 3.10 needs to use collections.abc.Iterable because typing.Iterable is removed
-try:
+if sys.version_info <= (3, 9):
     from typing import Iterable
-except ImportError:
+else:
     from collections.abc import Iterable
 
 logger = getLogger(__name__)
@@ -72,6 +77,15 @@ _MAX_INLINE_CLOSURE_SIZE_BYTES = 8192
 
 # Every table function handler class must define the process method.
 TABLE_FUNCTION_PROCESS_METHOD = "process"
+TABLE_FUNCTION_END_PARTITION_METHOD = "end_partition"
+
+# Every aggregate function handler class must define accumulate and finish methods.
+AGGREGATE_FUNCTION_ACCULUMATE_METHOD = "accumulate"
+AGGREGATE_FUNCTION_FINISH_METHOD = "finish"
+AGGREGATE_FUNCTION_MERGE_METHOD = "merge"
+AGGREGATE_FUNCTION_STATE_METHOD = "aggregate_state"
+
+EXECUTE_AS_WHITELIST = frozenset(["owner", "caller"])
 
 
 class UDFColumn(NamedTuple):
@@ -83,42 +97,205 @@ def is_local_python_file(file_path: str) -> bool:
     return not file_path.startswith(STAGE_PREFIX) and file_path.endswith(".py")
 
 
+def get_python_types_dict_for_udaf(
+    accumulate_hints: Dict[str, Any], finish_hints: Dict[str, Any]
+) -> Dict[str, Any]:
+    python_types_dict = {k: v for k, v in accumulate_hints.items() if k != "return"}
+    if "return" in finish_hints:
+        python_types_dict["return"] = finish_hints["return"]
+    return python_types_dict
+
+
+def get_python_types_dict_for_udtf(
+    process: Dict[str, Any], end_partition: Dict[str, Any]
+) -> Dict[str, Any]:
+    # Prefer input types from process and return types from end_partition
+    python_types_dict = {**end_partition, **process}
+    if "return" in end_partition:
+        python_types_dict["return"] = end_partition["return"]
+    return python_types_dict
+
+
+def extract_return_type_from_udtf_type_hints(
+    return_type_hint, output_schema, func_name
+) -> Union[StructType, "PandasDataFrameType", None]:
+    if return_type_hint is None and output_schema is not None:
+        raise ValueError(
+            "The return type hint is not set but 'output_schema' has only column names. You can either use a StructType instance for 'output_schema', or use"
+            "a combination of a return type hint for method 'process' and column names for 'output_schema'."
+        )
+    if typing.get_origin(return_type_hint) in [
+        list,
+        tuple,
+        collections.abc.Iterable,
+        collections.abc.Iterator,
+    ]:
+        row_type_hint = typing.get_args(return_type_hint)[0]  # The inner Tuple
+        if typing.get_origin(row_type_hint) != tuple:
+            raise ValueError(
+                f"The return type hint of method '{func_name}.process' must be a collection of tuples, for instance, Iterable[Tuple[str, int]], if you specify return type hint."
+            )
+        column_type_hints = typing.get_args(row_type_hint)
+        if len(column_type_hints) > 1 and column_type_hints[1] == Ellipsis:
+            return StructType(
+                [
+                    StructField(
+                        name,
+                        type_utils.python_type_to_snow_type(column_type_hints[0])[0],
+                    )
+                    for name in output_schema
+                ]
+            )
+        elif output_schema:
+            if len(column_type_hints) != len(output_schema):
+                raise ValueError(
+                    f"'output_schema' has {len(output_schema)} names while type hints Tuple has only {len(column_type_hints)}."
+                )
+            return StructType(
+                [
+                    StructField(
+                        name,
+                        type_utils.python_type_to_snow_type(column_type)[0],
+                    )
+                    for name, column_type in zip(output_schema, column_type_hints)
+                ]
+            )
+        else:  # both type hints and return type are specified
+            return None
+    elif return_type_hint is None:
+        return None
+    else:
+        if installed_pandas:  # Vectorized UDTF
+            if typing.get_origin(return_type_hint) == PandasDataFrame:
+                return PandasDataFrameType(
+                    col_types=[
+                        python_type_to_snow_type(x)[0]
+                        for x in typing.get_args(return_type_hint)
+                    ],
+                    col_names=output_schema,
+                )
+            elif return_type_hint is pandas.DataFrame:
+                return PandasDataFrameType(
+                    []
+                )  # placeholder, indicating the return type is pandas DataFrame
+        if return_type_hint is NoneType:
+            return None
+        else:
+            raise ValueError(
+                f"The return type hint for a UDTF handler must be a collection type or None or a PandasDataFrame. {return_type_hint} is used."
+            )
+
+
 def get_types_from_type_hints(
     func: Union[Callable, Tuple[str, str]],
     object_type: TempObjectType,
+    output_schema: Optional[List[str]] = None,
 ) -> Tuple[DataType, List[DataType]]:
     if isinstance(func, Callable):
         # For Python 3.10+, the result values of get_type_hints()
         # will become strings, which we have to change the implementation
         # here at that time. https://www.python.org/dev/peps/pep-0563/
-        python_types_dict = get_type_hints(
-            getattr(func, TABLE_FUNCTION_PROCESS_METHOD, func)
-        )
+        func_name = func.__name__
+        try:
+            if object_type == TempObjectType.AGGREGATE_FUNCTION:
+                accumulate_hints = get_type_hints(
+                    getattr(func, AGGREGATE_FUNCTION_ACCULUMATE_METHOD, func)
+                )
+                finish_hints = get_type_hints(
+                    getattr(func, AGGREGATE_FUNCTION_FINISH_METHOD, func)
+                )
+                python_types_dict = get_python_types_dict_for_udaf(
+                    accumulate_hints, finish_hints
+                )
+
+            elif object_type == TempObjectType.TABLE_FUNCTION:
+                if not (
+                    hasattr(func, TABLE_FUNCTION_END_PARTITION_METHOD)
+                    or hasattr(func, TABLE_FUNCTION_PROCESS_METHOD)
+                ):
+                    raise AttributeError(
+                        f"Neither `{TABLE_FUNCTION_PROCESS_METHOD}` nor `{TABLE_FUNCTION_END_PARTITION_METHOD}` is defined for class {func}"
+                    )
+                process_types_dict = {}
+                end_partition_types_dict = {}
+                # PROCESS and END_PARTITION have the same return type but input types might be different, favor PROCESS's types if both methods are present
+                if hasattr(func, TABLE_FUNCTION_PROCESS_METHOD):
+                    process_types_dict = get_type_hints(
+                        getattr(func, TABLE_FUNCTION_PROCESS_METHOD)
+                    )
+                if hasattr(func, TABLE_FUNCTION_END_PARTITION_METHOD):
+                    end_partition_types_dict = get_type_hints(
+                        getattr(func, TABLE_FUNCTION_END_PARTITION_METHOD)
+                    )
+                python_types_dict = get_python_types_dict_for_udtf(
+                    process_types_dict, end_partition_types_dict
+                )
+            else:
+                python_types_dict = get_type_hints(func)
+        except TypeError:
+            # if we fail to run get_type_hints on a function (a TypeError will be raised),
+            # return empty type dict. This will fail for functions like numpy.ufunc
+            # (e.g., get_type_hints(np.exp))
+            python_types_dict = {}
+        return_type_hint = python_types_dict.get("return")
     else:
-        if object_type == TempObjectType.TABLE_FUNCTION:
-            python_types_dict = (
-                retrieve_func_type_hints_from_source(
-                    func[0], TABLE_FUNCTION_PROCESS_METHOD, class_name=func[1]
-                )  # use method process of a UDTF handler class.
-                if is_local_python_file(func[0])
-                else {}
+        # Register from file
+        filename, func_name = func[0], func[1]
+        if not is_local_python_file(filename):
+            python_types_dict = {}
+        elif object_type == TempObjectType.AGGREGATE_FUNCTION:
+            accumulate_hints = retrieve_func_type_hints_from_source(
+                filename, AGGREGATE_FUNCTION_ACCULUMATE_METHOD, class_name=func_name
+            )
+            finish_hints = retrieve_func_type_hints_from_source(
+                filename, AGGREGATE_FUNCTION_FINISH_METHOD, class_name=func_name
+            )
+            python_types_dict = get_python_types_dict_for_udaf(
+                accumulate_hints, finish_hints
+            )
+        elif object_type == TempObjectType.TABLE_FUNCTION:
+            process_types_dict = retrieve_func_type_hints_from_source(
+                filename, TABLE_FUNCTION_PROCESS_METHOD, class_name=func[1]
+            )
+            end_partition_types_dict = retrieve_func_type_hints_from_source(
+                func[0], TABLE_FUNCTION_END_PARTITION_METHOD, class_name=func[1]
+            )
+            if process_types_dict is None and end_partition_types_dict is None:
+                raise ValueError(
+                    f"Neither {func_name}.{TABLE_FUNCTION_PROCESS_METHOD} or {func_name}.{TABLE_FUNCTION_END_PARTITION_METHOD} could be found from {filename}"
+                )
+            python_types_dict = get_python_types_dict_for_udtf(
+                process_types_dict or {}, end_partition_types_dict or {}
+            )
+        elif object_type in (TempObjectType.FUNCTION, TempObjectType.PROCEDURE):
+            python_types_dict = retrieve_func_type_hints_from_source(
+                filename, func_name
             )
         else:
-            python_types_dict = (
-                retrieve_func_type_hints_from_source(func[0], func[1])
-                if is_local_python_file(func[0])
-                else {}
+            raise ValueError(
+                f"Expecting FUNCTION, PROCEDURE, TABLE_FUNCTION, or AGGREGATE_FUNCTION as object_type, got {object_type}"
             )
 
+        if "return" in python_types_dict:
+            return_type_hint = python_type_str_to_object(
+                python_types_dict["return"], object_type == TempObjectType.PROCEDURE
+            )
+        else:
+            return_type_hint = None
+
     if object_type == TempObjectType.TABLE_FUNCTION:
-        return_type = None
-        # The return type is processed in udtf.py. Return None here.
+        return_type = extract_return_type_from_udtf_type_hints(
+            return_type_hint, output_schema, func_name
+        )
     else:
         return_type = (
-            python_type_to_snow_type(python_types_dict["return"])[0]
+            python_type_to_snow_type(
+                python_types_dict["return"], object_type == TempObjectType.PROCEDURE
+            )[0]
             if "return" in python_types_dict
             else None
         )
+
     input_types = []
 
     # types are in order
@@ -134,7 +311,11 @@ def get_types_from_type_hints(
                     "The first argument of stored proc function should be Session"
                 )
         elif key != "return":
-            input_types.append(python_type_to_snow_type(python_type)[0])
+            input_types.append(
+                python_type_to_snow_type(
+                    python_type, object_type == TempObjectType.PROCEDURE
+                )[0]
+            )
         index += 1
 
     return return_type, input_types
@@ -147,6 +328,8 @@ def get_error_message_abbr(object_type: TempObjectType) -> str:
         return "stored proc"
     if object_type == TempObjectType.TABLE_FUNCTION:
         return "table function"
+    if object_type == TempObjectType.AGGREGATE_FUNCTION:
+        return "aggregate function"
     raise ValueError(f"Expect FUNCTION of PROCEDURE, but get {object_type}")
 
 
@@ -173,6 +356,30 @@ def check_register_args(
         )
 
 
+def check_execute_as_arg(execute_as: typing.Literal["caller", "owner"]):
+    if (
+        not isinstance(execute_as, str)
+        or execute_as.lower() not in EXECUTE_AS_WHITELIST
+    ):
+        raise TypeError(
+            f"'execute_as' value '{execute_as}' is invalid, choose from "
+            f"{', '.join(EXECUTE_AS_WHITELIST, )}"
+        )
+
+
+def check_python_runtime_version(runtime_version_from_requirement: Optional[str]):
+    system_version = f"{sys.version_info[0]}.{sys.version_info[1]}"
+    if (
+        runtime_version_from_requirement is not None
+        and runtime_version_from_requirement != system_version
+    ):
+        raise ValueError(
+            f"Cloudpickle can only be used to send objects between the exact same version of Python. "
+            f"Your system version is {system_version} while your requirements have specified version "
+            f"{runtime_version_from_requirement}!"
+        )
+
+
 def process_file_path(file_path: str) -> str:
     file_path = file_path.strip()
     if not file_path.startswith(STAGE_PREFIX) and not os.path.exists(file_path):
@@ -185,21 +392,30 @@ def extract_return_input_types(
     return_type: Optional[DataType],
     input_types: Optional[List[DataType]],
     object_type: TempObjectType,
-) -> Tuple[bool, bool, DataType, List[DataType]]:
-    # return results are: is_pandas_udf, is_dataframe_input, return_type, input_types
+    output_schema: Optional[List[str]] = None,
+) -> Tuple[bool, bool, Union[DataType, List[DataType]], List[DataType]]:
+    """
+    Returns:
+        is_pandas_udf
+        is_dataframe_input
+        return_types
+        input_types
+
+    Notes:
+        There are 3 cases:
+           1. return_type and input_types are provided:
+              a. type hints are provided and they are all pandas.Series or pandas.DataFrame,
+                 then combine them to pandas-related types.
+              b. otherwise, just use return_type and input_types.
+           2. return_type and input_types are not provided, but type hints are provided,
+              then just use the types inferred from type hints.
+    """
 
-    # there are 3 cases:
-    #   1. return_type and input_types are provided:
-    #      a. type hints are provided and they are all pandas.Series or pandas.DataFrame,
-    #         then combine them to pandas-related types.
-    #      b. otherwise, just use return_type and input_types.
-    #   2. return_type and input_types are not provided, but type hints are provided,
-    #      then just use the types inferred from type hints.
     (
         return_type_from_type_hints,
         input_types_from_type_hints,
-    ) = get_types_from_type_hints(func, object_type)
-    if return_type and return_type_from_type_hints:
+    ) = get_types_from_type_hints(func, object_type, output_schema)
+    if installed_pandas and return_type and return_type_from_type_hints:
         if isinstance(return_type_from_type_hints, PandasSeriesType):
             res_return_type = (
                 return_type.element_type
@@ -226,12 +442,19 @@ def extract_return_input_types(
                 isinstance(tp, PandasSeriesType) for tp in input_types_from_type_hints
             ):
                 return True, False, res_return_type, res_input_types
+        elif isinstance(
+            return_type_from_type_hints, PandasDataFrameType
+        ):  # vectorized UDTF
+            return_type = PandasDataFrameType(
+                [x.datatype for x in return_type], [x.name for x in return_type]
+            )
 
     res_return_type = return_type or return_type_from_type_hints
     res_input_types = input_types or input_types_from_type_hints
 
     if not res_return_type or (
-        isinstance(res_return_type, PandasSeriesType)
+        installed_pandas
+        and isinstance(res_return_type, PandasSeriesType)
         and not res_return_type.element_type
     ):
         raise TypeError("The return type must be specified")
@@ -256,6 +479,9 @@ def extract_return_input_types(
                 f"the number of argument type hints ({len(input_types_from_type_hints)})"
             )
 
+    if not installed_pandas:
+        return False, False, res_return_type, res_input_types
+
     if isinstance(res_return_type, PandasSeriesType):
         if len(res_input_types) == 0:
             return True, False, res_return_type.element_type, []
@@ -266,10 +492,7 @@ def extract_return_input_types(
                 True,
                 True,
                 res_return_type.element_type,
-                [
-                    tp.element_type if isinstance(tp, PandasSeriesType) else tp
-                    for tp in res_input_types[0].col_types
-                ],
+                res_input_types[0].get_snowflake_col_datatypes(),
             )
         elif all(isinstance(tp, PandasSeriesType) for tp in res_input_types):
             return (
@@ -278,6 +501,20 @@ def extract_return_input_types(
                 res_return_type.element_type,
                 [tp.element_type for tp in res_input_types],
             )
+    elif isinstance(res_return_type, PandasDataFrameType):
+        if len(res_input_types) == 0:
+            return True, True, res_return_type, []
+        elif len(res_input_types) == 1 and isinstance(
+            res_input_types[0], PandasDataFrameType
+        ):
+            return (
+                True,
+                True,
+                res_return_type,
+                res_input_types[0].get_snowflake_col_datatypes(),
+            )
+        else:
+            return True, True, res_return_type, res_input_types
 
     # not pandas UDF
     if not isinstance(res_return_type, (PandasSeriesType, PandasDataFrameType)) and all(
@@ -287,7 +524,7 @@ def extract_return_input_types(
         return False, False, res_return_type, res_input_types
 
     raise TypeError(
-        f"Invalid return type or input types for UDF: return type {res_return_type}, input types {res_input_types}"
+        f"Invalid return type or input types: return type {res_return_type}, input types {res_input_types}"
     )
 
 
@@ -299,8 +536,13 @@ def process_registration_inputs(
     input_types: Optional[List[DataType]],
     name: Optional[Union[str, Iterable[str]]],
     anonymous: bool = False,
+    output_schema: Optional[List[str]] = None,
 ) -> Tuple[str, bool, bool, DataType, List[DataType]]:
-    # get the udf name
+    """
+
+    Args:
+        output_schema: List of column names of in the output, only applicable to UDTF
+    """
     if name:
         object_name = name if isinstance(name, str) else ".".join(name)
     else:
@@ -318,7 +560,7 @@ def process_registration_inputs(
         return_type,
         input_types,
     ) = extract_return_input_types(
-        func, return_type, [] if input_types is None else input_types, object_type
+        func, return_type, input_types or [], object_type, output_schema
     )
 
     return object_name, is_pandas_udf, is_dataframe_input, return_type, input_types
@@ -376,22 +618,44 @@ def generate_python_code(
     # annotations. However, we still serialize the original method because the extracted
     # function will have an extra argument `cls` or `self` from the class.
     if object_type == TempObjectType.TABLE_FUNCTION:
-        target_func = getattr(func, TABLE_FUNCTION_PROCESS_METHOD)
+        if is_pandas_udf:
+            annotated_funcs = [getattr(func, TABLE_FUNCTION_END_PARTITION_METHOD)]
+        else:
+            annotated_funcs = [getattr(func, TABLE_FUNCTION_PROCESS_METHOD)]
+    elif object_type == TempObjectType.AGGREGATE_FUNCTION:
+        annotated_funcs = [
+            getattr(func, AGGREGATE_FUNCTION_ACCULUMATE_METHOD),
+            getattr(func, AGGREGATE_FUNCTION_FINISH_METHOD),
+            getattr(func, AGGREGATE_FUNCTION_MERGE_METHOD),
+            getattr(func, AGGREGATE_FUNCTION_STATE_METHOD).fget,
+        ]
+    elif object_type in (TempObjectType.FUNCTION, TempObjectType.PROCEDURE):
+        annotated_funcs = [getattr(func, "__func__", func)]
     else:
-        target_func = getattr(func, "__func__", func)
+        raise ValueError(
+            f"Expecting FUNCTION, PROCEDURE, TABLE_FUNCTION or AGGREGATE_FUNCTION for object_type, got {object_type}"
+        )
 
     # clear the annotations because when the user annotates Variant and Geography,
     # which are from snowpark modules and will not work on the server side
     # built-in functions don't have __annotations__
-    if hasattr(target_func, "__annotations__"):
-        annotations = target_func.__annotations__
+    # TODO(SNOW-861329): This removal could cause race condition if two sessions are trying to register UDF using the
+    #   same decorated function/class.
+    annotations = [
+        getattr(annotated_func, "__annotations__", None)
+        for annotated_func in annotated_funcs
+    ]
+    if any(annotations):
         try:
-            target_func.__annotations__ = {}
+            for annotated_func in annotated_funcs:
+                annotated_func.__annotations__ = {}
             # we still serialize the original function
             pickled_func = pickle_function(func)
         finally:
             # restore the annotations so we don't change the original function
-            target_func.__annotations__ = annotations
+            for annotated_func, annotation in zip(annotated_funcs, annotations):
+                if annotation:
+                    annotated_func.__annotations__ = annotation
     else:
         pickled_func = pickle_function(func)
     args = ",".join(arg_names)
@@ -422,6 +686,17 @@ def {_DEFAULT_HANDLER_NAME}({args}):
     return func({args})
 """
     else:
+        # UDF and UDTF use wrapper functions to invoke the pickled functions.
+        if is_pandas_udf:
+            func_args = "df"
+            wrapper_params = func_args
+            if not is_dataframe_input:
+                func_args = "*[df[idx] for idx in range(df.shape[1])]"
+        else:
+            func_args = args
+            wrapper_params = args
+
+        # Function/Class definition
         func_code = """
 
 from threading import RLock
@@ -454,50 +729,60 @@ end_partition_invoked = InvokedFlag()
 class {_DEFAULT_HANDLER_NAME}(func):
     def __init__(self):
         lock_function_once(super().__init__, init_invoked)()
-
-    def process(self, {args}):
-        return lock_function_once(super().process, process_invoked)({args})
 """
-            if hasattr(func, "end_partition"):
+            if hasattr(func, TABLE_FUNCTION_PROCESS_METHOD):
                 func_code = f"""{func_code}
-    def end_partition(self):
-        return lock_function_once(super().end_partition, end_partition_invoked)()
+    def process(self, {wrapper_params}):
+        return lock_function_once(super().process, process_invoked)({func_args})
 """
-        elif is_pandas_udf:
+            if hasattr(func, TABLE_FUNCTION_END_PARTITION_METHOD):
+                func_code = f"""{func_code}
+    def end_partition(self, {wrapper_params if is_pandas_udf else ""}):
+        return lock_function_once(super().end_partition, end_partition_invoked)({func_args if is_pandas_udf else ""})
+"""
+        elif object_type == TempObjectType.AGGREGATE_FUNCTION:
+            func_code = f"""{func_code}
+init_invoked = InvokedFlag()
+accumulate_invoked = InvokedFlag()
+merge_invoked = InvokedFlag()
+finish_invoked = InvokedFlag()
+
+class {_DEFAULT_HANDLER_NAME}(func):
+    def __init__(self):
+        lock_function_once(super().__init__, init_invoked)()
+
+    def accumulate(self, {args}):
+        return lock_function_once(super().accumulate, accumulate_invoked)({args})
+
+    def merge(self, other_agg_state):
+        return lock_function_once(super().merge, merge_invoked)(other_agg_state)
+
+    def finish(self):
+        return lock_function_once(super().finish, finish_invoked)()
+            """
+        elif object_type == TempObjectType.FUNCTION:
+            func_code = f"""{func_code}
+invoked = InvokedFlag()
+
+def {_DEFAULT_HANDLER_NAME}({wrapper_params}):
+    return lock_function_once(func, invoked)({func_args})
+""".rstrip()
+
+        # Vectorized UDxF attributes
+        if is_pandas_udf:
             pandas_code = f"""
 import pandas
 
-{_DEFAULT_HANDLER_NAME}._sf_vectorized_input = pandas.DataFrame
+{_DEFAULT_HANDLER_NAME}{("."+TABLE_FUNCTION_END_PARTITION_METHOD) if object_type == TempObjectType.TABLE_FUNCTION else ""}._sf_vectorized_input = pandas.DataFrame
 """.rstrip()
             if max_batch_size:
                 pandas_code = f"""
 {pandas_code}
 {_DEFAULT_HANDLER_NAME}._sf_max_batch_size = {int(max_batch_size)}
-""".rstrip()
-            if is_dataframe_input:
-                func_code = f"""{func_code}
-invoked = InvokedFlag()
-
-def {_DEFAULT_HANDLER_NAME}(df):
-    return lock_function_once(func, invoked)(df)
-""".rstrip()
-            else:
-                func_code = f"""{func_code}
-invoked = InvokedFlag()
-
-def {_DEFAULT_HANDLER_NAME}(df):
-    return lock_function_once(func, invoked)(*[df[idx] for idx in range(df.shape[1])])
 """.rstrip()
             func_code = f"""
 {func_code}
 {pandas_code}
-""".rstrip()
-        else:
-            func_code = f"""{func_code}
-invoked = InvokedFlag()
-
-def {_DEFAULT_HANDLER_NAME}({args}):
-    return lock_function_once(func, invoked)({args})
 """.rstrip()
 
     return f"""
@@ -522,13 +807,32 @@ def resolve_imports_and_packages(
     *,
     statement_params: Optional[Dict[str, str]] = None,
     source_code_display: bool = False,
-) -> Tuple[str, str, str, str, str]:
-    upload_stage = (
+    skip_upload_on_content_match: bool = False,
+    is_permanent: bool = False,
+    force_inline_code: bool = False,
+) -> Tuple[str, str, str, str, str, bool]:
+    import_only_stage = (
         unwrap_stage_location_single_quote(stage_location)
         if stage_location
         else session.get_session_stage()
     )
 
+    upload_and_import_stage = (
+        import_only_stage if is_permanent else session.get_session_stage()
+    )
+
+    # resolve packages
+    resolved_packages = (
+        session._resolve_packages(packages, include_pandas=is_pandas_udf)
+        if packages is not None
+        else session._resolve_packages(
+            [],
+            session._packages,
+            validate_package=False,
+            include_pandas=is_pandas_udf,
+        )
+    )
+
     # resolve imports
     if imports:
         udf_level_imports = {}
@@ -546,28 +850,28 @@ def resolve_imports_and_packages(
                 )
             udf_level_imports[resolved_import_tuple[0]] = resolved_import_tuple[1:]
         all_urls = session._resolve_imports(
-            upload_stage, udf_level_imports, statement_params=statement_params
+            import_only_stage,
+            upload_and_import_stage,
+            udf_level_imports,
+            statement_params=statement_params,
         )
     elif imports is None:
         all_urls = session._resolve_imports(
-            upload_stage, statement_params=statement_params
+            import_only_stage,
+            upload_and_import_stage,
+            statement_params=statement_params,
         )
     else:
         all_urls = []
 
-    # resolve packages
-    resolved_packages = (
-        session._resolve_packages(packages, include_pandas=is_pandas_udf)
-        if packages is not None
-        else session._resolve_packages(
-            [], session._packages, validate_package=False, include_pandas=is_pandas_udf
-        )
-    )
-
     dest_prefix = get_udf_upload_prefix(udf_name)
 
     # Upload closure to stage if it is beyond inline closure size limit
     if isinstance(func, Callable):
+        custom_python_runtime_version_allowed = (
+            False  # As cloudpickle is being used, we cannot allow a custom runtime
+        )
+
         # generate a random name for udf py file
         # and we compress it first then upload it
         udf_file_name_base = f"udf_py_{random_number()}"
@@ -581,10 +885,10 @@ def resolve_imports_and_packages(
             max_batch_size,
             source_code_display=source_code_display,
         )
-        if len(code) > _MAX_INLINE_CLOSURE_SIZE_BYTES:
+        if not force_inline_code and len(code) > _MAX_INLINE_CLOSURE_SIZE_BYTES:
             dest_prefix = get_udf_upload_prefix(udf_name)
             upload_file_stage_location = normalize_remote_file_or_dir(
-                f"{upload_stage}/{dest_prefix}/{udf_file_name}"
+                f"{upload_and_import_stage}/{dest_prefix}/{udf_file_name}"
             )
             udf_file_name_base = os.path.splitext(udf_file_name)[0]
             with io.BytesIO() as input_stream:
@@ -594,7 +898,7 @@ def resolve_imports_and_packages(
                     zf.writestr(f"{udf_file_name_base}.py", code)
                 session._conn.upload_stream(
                     input_stream=input_stream,
-                    stage_location=upload_stage,
+                    stage_location=upload_and_import_stage,
                     dest_filename=udf_file_name,
                     dest_prefix=dest_prefix,
                     parallel=parallel,
@@ -602,6 +906,7 @@ def resolve_imports_and_packages(
                     compress_data=False,
                     overwrite=True,
                     is_in_udf=True,
+                    skip_upload_on_content_match=skip_upload_on_content_match,
                 )
             all_urls.append(upload_file_stage_location)
             inline_code = None
@@ -611,6 +916,7 @@ def resolve_imports_and_packages(
             upload_file_stage_location = None
             handler = _DEFAULT_HANDLER_NAME
     else:
+        custom_python_runtime_version_allowed = True
         udf_file_name = os.path.basename(func[0])
         # for a compressed file, it might have multiple extensions
         # and we should remove all extensions
@@ -623,15 +929,16 @@ def resolve_imports_and_packages(
             all_urls.append(func[0])
         else:
             upload_file_stage_location = normalize_remote_file_or_dir(
-                f"{upload_stage}/{dest_prefix}/{udf_file_name}"
+                f"{upload_and_import_stage}/{dest_prefix}/{udf_file_name}"
             )
             session._conn.upload_file(
                 path=func[0],
-                stage_location=upload_stage,
+                stage_location=upload_and_import_stage,
                 dest_prefix=dest_prefix,
                 parallel=parallel,
                 compress_data=False,
                 overwrite=True,
+                skip_upload_on_content_match=skip_upload_on_content_match,
             )
             all_urls.append(upload_file_stage_location)
 
@@ -640,7 +947,14 @@ def resolve_imports_and_packages(
         [url if is_single_quoted(url) else f"'{url}'" for url in all_urls]
     )
     all_packages = ",".join([f"'{package}'" for package in resolved_packages])
-    return handler, inline_code, all_imports, all_packages, upload_file_stage_location
+    return (
+        handler,
+        inline_code,
+        all_imports,
+        all_packages,
+        upload_file_stage_location,
+        custom_python_runtime_version_allowed,
+    )
 
 
 def create_python_udf_or_sp(
@@ -652,7 +966,7 @@ def create_python_udf_or_sp(
     object_name: str,
     all_imports: str,
     all_packages: str,
-    is_temporary: bool,
+    is_permanent: bool,
     replace: bool,
     if_not_exists: bool,
     inline_python_code: Optional[str] = None,
@@ -660,11 +974,21 @@ def create_python_udf_or_sp(
     api_call_source: Optional[str] = None,
     strict: bool = False,
     secure: bool = False,
+    external_access_integrations: Optional[List[str]] = None,
+    secrets: Optional[Dict[str, str]] = None,
+    immutable: bool = False,
 ) -> None:
+    runtime_version = (
+        f"{sys.version_info[0]}.{sys.version_info[1]}"
+        if not session._runtime_version_from_requirement
+        else session._runtime_version_from_requirement
+    )
     if replace and if_not_exists:
         raise ValueError("options replace and if_not_exists are incompatible")
     if isinstance(return_type, StructType):
         return_sql = f'RETURNS TABLE ({",".join(f"{field.name} {convert_sp_to_sf_type(field.datatype)}" for field in return_type.fields)})'
+    elif installed_pandas and isinstance(return_type, PandasDataFrameType):
+        return_sql = f'RETURNS TABLE ({",".join(f"{name} {convert_sp_to_sf_type(datatype)}" for name, datatype in zip(return_type.col_names, return_type.col_types))})'
     else:
         return_sql = f"RETURNS {convert_sp_to_sf_type(return_type)}"
     input_sql_types = [convert_sp_to_sf_type(arg.datatype) for arg in input_args]
@@ -692,21 +1016,35 @@ $$
         if inline_python_code
         else ""
     )
-
+    mutability = "IMMUTABLE" if immutable else "VOLATILE"
     strict_as_sql = "\nSTRICT" if strict else ""
 
+    external_access_integrations_in_sql = (
+        f"\nEXTERNAL_ACCESS_INTEGRATIONS=({','.join(external_access_integrations)})"
+        if external_access_integrations
+        else ""
+    )
+    secrets_in_sql = (
+        f"""\nSECRETS=({",".join([f"'{k}'={v}" for k, v in secrets.items()])})"""
+        if secrets
+        else ""
+    )
+
     create_query = f"""
 CREATE{" OR REPLACE " if replace else ""}
-{"TEMPORARY" if is_temporary else ""} {"SECURE" if secure else ""} {object_type.value} {"IF NOT EXISTS" if if_not_exists else ""} {object_name}({sql_func_args})
+{"" if is_permanent else "TEMPORARY"} {"SECURE" if secure else ""} {object_type.value.replace("_", " ")} {"IF NOT EXISTS" if if_not_exists else ""} {object_name}({sql_func_args})
 {return_sql}
 LANGUAGE PYTHON {strict_as_sql}
-RUNTIME_VERSION={sys.version_info[0]}.{sys.version_info[1]}
+{mutability}
+RUNTIME_VERSION={runtime_version}
 {imports_in_sql}
 {packages_in_sql}
+{external_access_integrations_in_sql}
+{secrets_in_sql}
 HANDLER='{handler}'{execute_as_sql}
 {inline_python_code_in_sql}
 """
-    session._run_query(create_query, is_ddl_on_temp_object=is_temporary)
+    session._run_query(create_query, is_ddl_on_temp_object=not is_permanent)
 
     # fire telemetry after _run_query is successful
     api_call_source = api_call_source or "_internal.create_python_udf_or_sp"
@@ -725,8 +1063,19 @@ def generate_anonymous_python_sp_sql(
     all_packages: str,
     inline_python_code: Optional[str] = None,
     strict: bool = False,
+    runtime_version: Optional[str] = None,
+    external_access_integrations: Optional[List[str]] = None,
+    secrets: Optional[Dict[str, str]] = None,
 ):
-    return_sql = f"RETURNS {convert_sp_to_sf_type(return_type)}"
+    runtime_version = (
+        f"{sys.version_info[0]}.{sys.version_info[1]}"
+        if not runtime_version
+        else runtime_version
+    )
+    if isinstance(return_type, StructType):
+        return_sql = f'RETURNS TABLE ({",".join(f"{field.name} {convert_sp_to_sf_type(field.datatype)}" for field in return_type.fields)})'
+    else:
+        return_sql = f"RETURNS {convert_sp_to_sf_type(return_type)}"
     input_sql_types = [convert_sp_to_sf_type(arg.datatype) for arg in input_args]
     sql_func_args = ",".join(
         [f"{a.name} {t}" for a, t in zip(input_args, input_sql_types)]
@@ -743,14 +1092,26 @@ $$
         else ""
     )
     strict_as_sql = "\nSTRICT" if strict else ""
+    external_access_integrations_in_sql = (
+        f"\nEXTERNAL_ACCESS_INTEGRATIONS=({','.join(external_access_integrations)})"
+        if external_access_integrations
+        else ""
+    )
+    secrets_in_sql = (
+        f"""\nSECRETS=({",".join([f"'{k}'={v}" for k, v in secrets.items()])})"""
+        if secrets
+        else ""
+    )
 
     sql = f"""
 WITH {object_name} AS PROCEDURE ({sql_func_args})
 {return_sql}
 LANGUAGE PYTHON {strict_as_sql}
-RUNTIME_VERSION={sys.version_info[0]}.{sys.version_info[1]}
+RUNTIME_VERSION={runtime_version}
 {imports_in_sql}
 {packages_in_sql}
+{external_access_integrations_in_sql}
+{secrets_in_sql}
 HANDLER='{handler}'
 {inline_python_code_in_sql}
 """
@@ -763,7 +1124,7 @@ def generate_call_python_sp_sql(
     sql_args = []
     for arg in args:
         if isinstance(arg, snowflake.snowpark.Column):
-            sql_args.append(session._analyzer.analyze(arg._expression))
+            sql_args.append(session._analyzer.analyze(arg._expression, {}))
         else:
             sql_args.append(to_sql(arg, infer_type(arg)))
     return f"CALL {sproc_name}({', '.join(sql_args)})"
