@@ -23,6 +23,7 @@ from json import JSONEncoder
 from random import choice
 from typing import (
     IO,
+    TYPE_CHECKING,
     Any,
     Callable,
     Dict,
@@ -32,6 +33,7 @@ from typing import (
     Optional,
     Tuple,
     Type,
+    Union,
 )
 
 import snowflake.snowpark
@@ -43,6 +45,12 @@ from snowflake.snowpark._internal.error_message import SnowparkClientExceptionMe
 from snowflake.snowpark.row import Row
 from snowflake.snowpark.version import VERSION as snowpark_version
 
+if TYPE_CHECKING:
+    try:
+        from snowflake.connector.cursor import ResultMetadataV2
+    except ImportError:
+        ResultMetadataV2 = ResultMetadata
+
 STAGE_PREFIX = "@"
 
 # Scala uses 3 but this can be larger. Consider allowing users to configure it.
@@ -85,6 +93,11 @@ SNOWFLAKE_SELECT_SQL_PREFIX_PATTERN = re.compile(
     r"^(\s|\()*(select|with)", re.IGNORECASE
 )
 
+# Anonymous stored procedures: https://docs.snowflake.com/en/sql-reference/sql/call-with
+SNOWFLAKE_ANONYMOUS_CALL_WITH_PATTERN = re.compile(
+    r"^\s*with\s+\w+\s+as\s+procedure", re.IGNORECASE
+)
+
 # A set of widely-used packages,
 # whose names in pypi are different from their package name
 PACKAGE_NAME_TO_MODULE_NAME_MAP = {
@@ -122,7 +135,7 @@ MODULE_NAME_TO_PACKAGE_NAME_MAP = {
 
 GENERATED_PY_FILE_EXT = (".pyc", ".pyo", ".pyd", ".pyi")
 
-INFER_SCHEMA_FORMAT_TYPES = ("PARQUET", "ORC", "AVRO")
+INFER_SCHEMA_FORMAT_TYPES = ("PARQUET", "ORC", "AVRO", "JSON", "CSV")
 
 COPY_OPTIONS = {
     "ON_ERROR",
@@ -173,6 +186,8 @@ class TempObjectType(Enum):
     COLUMN = "COLUMN"
     PROCEDURE = "PROCEDURE"
     TABLE_FUNCTION = "TABLE_FUNCTION"
+    DYNAMIC_TABLE = "DYNAMIC_TABLE"
+    AGGREGATE_FUNCTION = "AGGREGATE_FUNCTION"
 
 
 def validate_object_name(name: str):
@@ -224,7 +239,10 @@ def unwrap_single_quote(name: str) -> str:
 
 
 def is_sql_select_statement(sql: str) -> bool:
-    return SNOWFLAKE_SELECT_SQL_PREFIX_PATTERN.match(sql) is not None
+    return (
+        SNOWFLAKE_SELECT_SQL_PREFIX_PATTERN.match(sql) is not None
+        and SNOWFLAKE_ANONYMOUS_CALL_WITH_PATTERN.match(sql) is None
+    )
 
 
 def normalize_path(path: str, is_local: bool) -> str:
@@ -299,8 +317,8 @@ def zip_file_or_directory_to_stream(
             absolute path = [leading path]/[relative path]. For example,
             when the path is "/tmp/dir1/dir2/test.py", and the leading path
             is "/tmp/dir1", the generated filesystem structure in the zip file
-            will be "dir2/test.py".
-        add_init_py: Whether to add __init__.py along the compressed path.
+            will be "dir2/test.py". The leading path will compose a namespace package
+            that is used for zipimport on the server side.
         ignore_generated_py_file: Whether to ignore some generated python files
             in the directory.
 
@@ -319,6 +337,13 @@ def zip_file_or_directory_to_stream(
     with zipfile.ZipFile(
         input_stream, mode="w", compression=zipfile.ZIP_DEFLATED
     ) as zf:
+        # Write the folders on the leading path to the zip file to build a namespace package
+        cur_path = os.path.dirname(path)
+        while os.path.realpath(cur_path) != os.path.realpath(start_path):
+            # according to .zip file format specification, only / is valid
+            zf.writestr(f"{os.path.relpath(cur_path, start_path)}/", "")
+            cur_path = os.path.dirname(cur_path)
+
         if os.path.isdir(path):
             for dirname, _, files in os.walk(path):
                 # ignore __pycache__
@@ -336,15 +361,6 @@ def zip_file_or_directory_to_stream(
         else:
             zf.write(path, os.path.relpath(path, start_path))
 
-        # __init__.py is needed for all directories along the import path
-        # when importing a module as a zip file
-        if add_init_py:
-            relative_path = os.path.relpath(path, start_path)
-            head, _ = os.path.split(relative_path)
-            while head and head != os.sep:
-                zf.writestr(os.path.join(head, "__init__.py"), "")
-                head, _ = os.path.split(head)
-
     yield input_stream
     input_stream.close()
 
@@ -515,7 +531,7 @@ def column_to_bool(col_):
 
 def result_set_to_rows(
     result_set: List[Any],
-    result_meta: Optional[List[ResultMetadata]] = None,
+    result_meta: Optional[Union[List[ResultMetadata], List["ResultMetadataV2"]]] = None,
     case_sensitive: bool = True,
 ) -> List[Row]:
     col_names = [col.name for col in result_meta] if result_meta else None
@@ -616,6 +632,27 @@ def func_decorator(
     return wrapper
 
 
+def param_decorator(
+    decorator_type: Literal["deprecated", "experimental", "in private preview"],
+    *,
+    version: str,
+) -> Callable:
+    def wrapper(param_setter_function):
+        warning_text = (
+            f"Parameter {param_setter_function.__name__} is {decorator_type} since {version}. "
+            f"{'Do not use it in production. ' if decorator_type in ('experimental', 'in private preview') else ''}"
+        )
+
+        @functools.wraps(param_setter_function)
+        def func_call_wrapper(*args, **kwargs):
+            warning(param_setter_function.__name__, warning_text)
+            return param_setter_function(*args, **kwargs)
+
+        return func_call_wrapper
+
+    return wrapper
+
+
 def deprecated(
     *, version: str, extra_warning_text: str = "", extra_doc_string: str = ""
 ) -> Callable:
@@ -638,6 +675,13 @@ def experimental(
     )
 
 
+def experimental_parameter(*, version: str) -> Callable:
+    return param_decorator(
+        "experimental",
+        version=version,
+    )
+
+
 def private_preview(
     *, version: str, extra_warning_text: str = "", extra_doc_string: str = ""
 ) -> Callable:
@@ -679,3 +723,106 @@ def get_copy_into_table_options(
         elif k not in NON_FORMAT_TYPE_OPTIONS:
             file_format_type_options[k] = v
     return file_format_type_options, copy_options
+
+
+def strip_double_quotes_in_like_statement_in_table_name(table_name: str) -> str:
+    """
+    this function is used by method _table_exists to handle double quotes in table name when calling
+    SHOW TABLES LIKE
+    """
+    if not table_name or len(table_name) < 2:
+        return table_name
+
+    # escape double quotes, e.g. users pass """a.b""" as table name:
+    # df.write.save_as_table('"""a.b"""', mode="append")
+    # and we should call SHOW TABLES LIKE '"a.b"'
+    table_name = table_name.replace('""', '"')
+
+    # if table_name == '"a.b"', then we should call SHOW TABLES LIKE 'a.b'
+    return table_name[1:-1] if table_name[0] == table_name[-1] == '"' else table_name
+
+
+def parse_table_name(table_name: str) -> List[str]:
+    """
+    This function implements the algorithm to parse a table name.
+
+    We parse the table name according to the following rules:
+    https://docs.snowflake.com/en/sql-reference/identifiers-syntax
+
+    - Unquoted object identifiers:
+        - Start with a letter (A-Z, a-z) or an underscore (“_”).
+        - Contain only letters, underscores, decimal digits (0-9), and dollar signs (“$”).
+        - Are stored and resolved as uppercase characters (e.g. id is stored and resolved as ID).
+
+    - If you put double quotes around an identifier (e.g. “My identifier with blanks and punctuation.”),
+        the following rules apply:
+        - The case of the identifier is preserved when storing and resolving the identifier (e.g. "id" is
+            stored and resolved as id).
+        - The identifier can contain and start with ASCII, extended ASCII, and non-ASCII characters.
+    """
+    validate_object_name(table_name)
+    str_len = len(table_name)
+    ret = []
+
+    in_double_quotes = False
+    i = 0
+    cur_word_start_idx = 0
+
+    while i < str_len:
+        cur_char = table_name[i]
+        if cur_char == '"':
+            if in_double_quotes:
+                # we have to check whether this `"` is the ending of a double-quoted identifier
+                # or it's an escaping double quote
+                # to achieve this, we need to preload one more char
+                if i < str_len - 1 and table_name[i + 1] == '"':
+                    # two consecutive '"', this is an escaping double quotes
+                    # the pointer just keeps moving forward
+                    i += 1
+                else:
+                    # the double quotes indicates the ending of an identifier
+                    in_double_quotes = False
+                    # it should be followed by a '.' for splitting, or it should reach the end of the str
+            else:
+                # this is the beginning of another double-quoted identifier
+                in_double_quotes = True
+        elif cur_char == ".":
+            if not in_double_quotes:
+                # this dot is to split db.schema.database
+                # we concatenate the processed chars into a string
+                # and append the string to the return list, and set our cur_word_start_idx to position after the dot
+                ret.append(table_name[cur_word_start_idx:i])
+                cur_word_start_idx = i + 1
+            # else dot is part of the table name
+        # else cur_char is part of the name
+        i += 1
+
+    ret.append(table_name[cur_word_start_idx:i])
+    return ret
+
+
+EMPTY_STRING = ""
+DOUBLE_QUOTE = '"'
+# Quoted values may also include newlines, so '.' must match _everything_ within quotes
+ALREADY_QUOTED = re.compile('^(".+")$', re.DOTALL)
+UNQUOTED_CASE_INSENSITIVE = re.compile("^([_A-Za-z]+[_A-Za-z0-9$]*)$")
+
+
+def quote_name(name: str) -> str:
+    if ALREADY_QUOTED.match(name):
+        return validate_quoted_name(name)
+    elif UNQUOTED_CASE_INSENSITIVE.match(name):
+        return DOUBLE_QUOTE + escape_quotes(name.upper()) + DOUBLE_QUOTE
+    else:
+        return DOUBLE_QUOTE + escape_quotes(name) + DOUBLE_QUOTE
+
+
+def validate_quoted_name(name: str) -> str:
+    if DOUBLE_QUOTE in name[1:-1].replace(DOUBLE_QUOTE + DOUBLE_QUOTE, EMPTY_STRING):
+        raise SnowparkClientExceptionMessages.PLAN_ANALYZER_INVALID_IDENTIFIER(name)
+    else:
+        return name
+
+
+def escape_quotes(unescaped: str) -> str:
+    return unescaped.replace(DOUBLE_QUOTE, DOUBLE_QUOTE + DOUBLE_QUOTE)
