@@ -2,6 +2,7 @@
 # Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
 #
 
+import sys
 from typing import Dict, List, Literal, Optional, Union, overload
 
 import snowflake.snowpark  # for forward references of type hints
@@ -14,25 +15,26 @@ from snowflake.snowpark._internal.telemetry import (
     add_api_call,
     dfw_collect_api_telemetry,
 )
-from snowflake.snowpark._internal.type_utils import ColumnOrSqlExpr
+from snowflake.snowpark._internal.type_utils import ColumnOrName, ColumnOrSqlExpr
 from snowflake.snowpark._internal.utils import (
     SUPPORTED_TABLE_TYPES,
     normalize_remote_file_or_dir,
+    parse_table_name,
     str_to_enum,
     validate_object_name,
     warning,
 )
 from snowflake.snowpark.async_job import AsyncJob, _AsyncResultType
-from snowflake.snowpark.column import Column
+from snowflake.snowpark.column import Column, _to_col_if_str
 from snowflake.snowpark.functions import sql_expr
 from snowflake.snowpark.row import Row
 
 # Python 3.8 needs to use typing.Iterable because collections.abc.Iterable is not subscriptable
 # Python 3.9 can use both
 # Python 3.10 needs to use collections.abc.Iterable because typing.Iterable is removed
-try:
+if sys.version_info <= (3, 9):
     from typing import Iterable
-except ImportError:
+else:
     from collections.abc import Iterable
 
 
@@ -59,13 +61,13 @@ class DataFrameWriter:
         Args:
             save_mode: One of the following strings.
 
-                "append": Append data of this DataFrame to existing data.
+                "append": Append data of this DataFrame to the existing table. Creates a table if it does not exist.
 
-                "overwrite": Overwrite existing data.
+                "overwrite": Overwrite the existing table.
 
-                "errorifexists": Throw an exception if data already exists.
+                "errorifexists": Throw an exception if the table already exists.
 
-                "ignore": Ignore this operation if data already exists.
+                "ignore": Ignore this operation if the table already exists.
 
                 Default value is "errorifexists".
 
@@ -84,6 +86,7 @@ class DataFrameWriter:
         column_order: str = "index",
         create_temp_table: bool = False,
         table_type: Literal["", "temp", "temporary", "transient"] = "",
+        clustering_keys: Iterable[Column],
         statement_params: Optional[Dict[str, str]] = None,
         block: bool = True,
     ) -> None:
@@ -98,6 +101,7 @@ class DataFrameWriter:
         column_order: str = "index",
         create_temp_table: bool = False,
         table_type: Literal["", "temp", "temporary", "transient"] = "",
+        clustering_keys: Iterable[Column],
         statement_params: Optional[Dict[str, str]] = None,
         block: bool = False,
     ) -> AsyncJob:
@@ -112,24 +116,26 @@ class DataFrameWriter:
         column_order: str = "index",
         create_temp_table: bool = False,
         table_type: Literal["", "temp", "temporary", "transient"] = "",
+        clustering_keys: Optional[Iterable[ColumnOrName]] = None,
         statement_params: Optional[Dict[str, str]] = None,
         block: bool = True,
     ) -> Optional[AsyncJob]:
         """Writes the data to the specified table in a Snowflake database.
 
         Args:
-            table_name: A string or list of strings that specify the table name or fully-qualified object identifier
-                (database name, schema name, and table name).
+            table_name: A string or list of strings representing table name.
+                If input is a string, it represents the table name; if input is of type iterable of strings,
+                it represents the fully-qualified object identifier (database name, schema name, and table name).
             mode: One of the following values. When it's ``None`` or not provided,
                 the save mode set by :meth:`mode` is used.
 
-                "append": Append data of this DataFrame to existing data.
+                "append": Append data of this DataFrame to the existing table. Creates a table if it does not exist.
 
-                "overwrite": Overwrite existing data.
+                "overwrite": Overwrite the existing table.
 
-                "errorifexists": Throw an exception if data already exists.
+                "errorifexists": Throw an exception if the table already exists.
 
-                "ignore": Ignore this operation if data already exists.
+                "ignore": Ignore this operation if the table already exists.
 
             column_order: When ``mode`` is "append", data will be inserted into the target table by matching column sequence or column name. Default is "index". When ``mode`` is not "append", the ``column_order`` makes no difference.
 
@@ -140,8 +146,11 @@ class DataFrameWriter:
             table_type: The table type of table to be created. The supported values are: ``temp``, ``temporary``,
                         and ``transient``. An empty string means to create a permanent table. Learn more about table
                         types `here <https://docs.snowflake.com/en/user-guide/tables-temp-transient.html>`_.
+            clustering_keys: Specifies one or more columns or column expressions in the table as the clustering key.
+                See `Clustering Keys & Clustered Tables <https://docs.snowflake.com/en/user-guide/tables-clustering-keys#defining-a-clustering-key-for-a-table>`_
+                for more details.
             statement_params: Dictionary of statement level parameters to be set while executing this action.
-            block: (Experimental) A bool value indicating whether this function will wait until the result is available.
+            block: A bool value indicating whether this function will wait until the result is available.
                 When it is ``False``, this function executes the underlying queries of the dataframe
                 asynchronously and returns an :class:`AsyncJob`.
 
@@ -158,12 +167,6 @@ class DataFrameWriter:
             >>> session.table("my_transient_table").collect()
             [Row(A=1, B=2), Row(A=3, B=4)]
         """
-        if not block:
-            warning(
-                "save_as_table.block",
-                "block argument is experimental. Do not use it in production.",
-            )
-
         save_mode = (
             str_to_enum(mode.lower(), SaveMode, "'mode'") if mode else self._save_mode
         )
@@ -171,11 +174,22 @@ class DataFrameWriter:
             table_name if isinstance(table_name, str) else ".".join(table_name)
         )
         validate_object_name(full_table_name)
+        table_name = (
+            parse_table_name(table_name) if isinstance(table_name, str) else table_name
+        )
         if column_order is None or column_order.lower() not in ("name", "index"):
             raise ValueError("'column_order' must be either 'name' or 'index'")
         column_names = (
             self._dataframe.columns if column_order.lower() == "name" else None
         )
+        clustering_exprs = (
+            [
+                _to_col_if_str(col, "DataFrameWriter.save_as_table")._expression
+                for col in clustering_keys
+            ]
+            if clustering_keys
+            else []
+        )
 
         if create_temp_table:
             warning(
@@ -191,17 +205,18 @@ class DataFrameWriter:
             )
 
         create_table_logic_plan = SnowflakeCreateTable(
-            full_table_name,
+            table_name,
             column_names,
             save_mode,
             self._dataframe._plan,
             table_type,
+            clustering_exprs,
         )
         session = self._dataframe._session
         snowflake_plan = session._analyzer.resolve(create_table_logic_plan)
         result = session._conn.execute(
             snowflake_plan,
-            _statement_params=statement_params,
+            _statement_params=statement_params or self._dataframe._statement_params,
             block=block,
             data_type=_AsyncResultType.NO_RESULT,
         )
@@ -263,7 +278,7 @@ class DataFrameWriter:
             header: Specifies whether to include the table column headings in the output files.
             statement_params: Dictionary of statement level parameters to be set while executing this action.
             copy_options: The kwargs that are used to specify the copy options. Use the options documented in the `Copy Options <https://docs.snowflake.com/en/sql-reference/sql/copy-into-location.html#copy-options-copyoptions>`__.
-            block: (Experimental) A bool value indicating whether this function will wait until the result is available.
+            block: A bool value indicating whether this function will wait until the result is available.
                 When it is ``False``, this function executes the underlying queries of the dataframe
                 asynchronously and returns an :class:`AsyncJob`.
 
@@ -293,12 +308,6 @@ class DataFrameWriter:
             FIRST_NAME: [["John","Rick","Anthony"]]
             LAST_NAME: [["Berry","Berry","Davis"]]
         """
-        if not block:
-            warning(
-                "copy_into_location.block",
-                "block argument is experimental. Do not use it in production.",
-            )
-
         stage_location = normalize_remote_file_or_dir(location)
         if isinstance(partition_by, str):
             partition_by = sql_expr(partition_by)._expression
@@ -322,7 +331,8 @@ class DataFrameWriter:
         )
         add_api_call(df, "DataFrameWriter.copy_into_location")
         return df._internal_collect_with_tag(
-            statement_params=statement_params, block=block
+            statement_params=statement_params or self._dataframe._statement_params,
+            block=block,
         )
 
     saveAsTable = save_as_table
