@@ -2,15 +2,15 @@
 #
 # Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
 #
+from typing import Callable, Dict, Iterable, List, Tuple, Union
 
-import re
-from typing import Callable, Dict, List, Tuple, Union
-
+from snowflake.connector.options import pandas
 from snowflake.snowpark import functions
 from snowflake.snowpark._internal.analyzer.expression import (
     Expression,
     Literal,
     NamedExpression,
+    SnowflakeUDF,
     UnresolvedAttribute,
 )
 from snowflake.snowpark._internal.analyzer.grouping_set import (
@@ -18,10 +18,6 @@ from snowflake.snowpark._internal.analyzer.grouping_set import (
     GroupingSetsExpression,
     Rollup,
 )
-from snowflake.snowpark._internal.analyzer.select_statement import (
-    SelectSnowflakePlan,
-    SelectStatement,
-)
 from snowflake.snowpark._internal.analyzer.unary_expression import (
     Alias,
     UnresolvedAlias,
@@ -29,28 +25,20 @@ from snowflake.snowpark._internal.analyzer.unary_expression import (
 from snowflake.snowpark._internal.analyzer.unary_plan_node import Aggregate, Pivot
 from snowflake.snowpark._internal.error_message import SnowparkClientExceptionMessages
 from snowflake.snowpark._internal.telemetry import relational_group_df_api_usage
-from snowflake.snowpark._internal.type_utils import ColumnOrName
+from snowflake.snowpark._internal.type_utils import ColumnOrName, LiteralType
 from snowflake.snowpark._internal.utils import parse_positional_args_to_list
 from snowflake.snowpark.column import Column
 from snowflake.snowpark.dataframe import DataFrame
-
-INVALID_SF_IDENTIFIER_CHARS = re.compile("[^\\x20-\\x7E]")
-
-
-def _strip_invalid_sf_identifier_chars(identifier: str) -> str:
-    return INVALID_SF_IDENTIFIER_CHARS.sub("", identifier.replace('"', ""))
+from snowflake.snowpark.types import StructType
 
 
 def _alias(expr: Expression) -> NamedExpression:
     if isinstance(expr, UnresolvedAttribute):
         return UnresolvedAlias(expr)
-    elif isinstance(expr, NamedExpression):
+    elif isinstance(expr, (NamedExpression, SnowflakeUDF)):
         return expr
     else:
-        return Alias(
-            expr,
-            _strip_invalid_sf_identifier_chars(expr.sql.upper()),
-        )
+        return Alias(expr, expr.sql.upper().replace('"', ""))
 
 
 def _expr_to_func(expr: str, input_expr: Expression) -> Expression:
@@ -62,7 +50,7 @@ def _expr_to_func(expr: str, input_expr: Expression) -> Expression:
     elif lowered in ["count", "size"]:
         return functions.count(Column(input_expr))._expression
     else:
-        return functions.function(lowered)(input_expr)._expression
+        return functions.function(expr)(input_expr)._expression
 
 
 def _str_to_expr(expr: str) -> Callable:
@@ -181,6 +169,7 @@ class RelationalGroupedDataFrame:
             if len(agg_exprs) != 1:
                 raise SnowparkClientExceptionMessages.DF_PIVOT_ONLY_SUPPORT_ONE_AGG_EXPR()
             group_plan = Pivot(
+                self._grouping_exprs,
                 self._group_type.pivot_col,
                 self._group_type.values,
                 agg_exprs,
@@ -190,9 +179,9 @@ class RelationalGroupedDataFrame:
             raise TypeError(f"Wrong group by type {self._group_type}")
 
         if self._df._select_statement:
-            group_plan = SelectStatement(
-                from_=SelectSnowflakePlan(
-                    snowflake_plan=group_plan, analyzer=self._df._session._analyzer
+            group_plan = self._df._session._analyzer.create_select_statement(
+                from_=self._df._session._analyzer.create_select_snowflake_plan(
+                    group_plan, analyzer=self._df._session._analyzer
                 ),
                 analyzer=self._df._session._analyzer,
             )
@@ -263,6 +252,172 @@ class RelationalGroupedDataFrame:
 
         return self._to_df(agg_exprs)
 
+    def apply_in_pandas(
+        self,
+        func: Callable,
+        output_schema: StructType,
+        **kwargs,
+    ) -> DataFrame:
+        """Maps each grouped dataframe in to a pandas.DataFrame, applies the given function on
+        data of each grouped dataframe, and returns a pandas.DataFrame. Internally, a vectorized
+        UDTF with input ``func`` argument as the ``end_partition`` is registered and called. Additional
+        ``kwargs`` are accepted to specify arguments to register the UDTF. Group by clause used must be
+        column reference, not a general expression.
+
+        Requires ``pandas`` to be installed in the execution environment and declared as a dependency by either
+        specifying the keyword argument `packages=["pandas]` in this call or calling :meth:`~snowflake.snowpark.Session.add_packages` beforehand.
+
+        Args:
+            func: A Python native function that accepts a single input argument - a ``pandas.DataFrame``
+                object and returns a ``pandas.Dataframe``. It is used as input to ``end_partition`` in
+                a vectorized UDTF.
+            output_schema: A :class:`~snowflake.snowpark.types.StructType` instance that represents the
+                table function's output columns.
+            input_names: A list of strings that represents the table function's input column names. Optional,
+                if unspecified, default column names will be ARG1, ARG2, etc.
+            kwargs: Additional arguments to register the vectorized UDTF. See
+                :meth:`~snowflake.snowpark.udtf.UDTFRegistration.register` for all options.
+
+        Examples::
+            Call ``apply_in_pandas`` using temporary UDTF:
+
+                >>> import pandas as pd
+                >>> from snowflake.snowpark.types import StructType, StructField, StringType, FloatType
+                >>> def convert(pandas_df):
+                ...     return pandas_df.assign(TEMP_F = lambda x: x.TEMP_C * 9 / 5 + 32)
+                >>> df = session.createDataFrame([('SF', 21.0), ('SF', 17.5), ('SF', 24.0), ('NY', 30.9), ('NY', 33.6)],
+                ...         schema=['location', 'temp_c'])
+                >>> df.group_by("location").apply_in_pandas(convert,
+                ...     output_schema=StructType([StructField("location", StringType()),
+                ...                               StructField("temp_c", FloatType()),
+                ...                               StructField("temp_f", FloatType())])).order_by("temp_c").show()
+                ---------------------------------------------
+                |"LOCATION"  |"TEMP_C"  |"TEMP_F"           |
+                ---------------------------------------------
+                |SF          |17.5      |63.5               |
+                |SF          |21.0      |69.8               |
+                |SF          |24.0      |75.2               |
+                |NY          |30.9      |87.61999999999999  |
+                |NY          |33.6      |92.48              |
+                ---------------------------------------------
+                <BLANKLINE>
+
+            Call ``apply_in_pandas`` using permanent UDTF with replacing original UDTF:
+
+                >>> from snowflake.snowpark.types import IntegerType, DoubleType
+                >>> _ = session.sql("create or replace temp stage mystage").collect()
+                >>> def group_sum(pdf):
+                ...     return pd.DataFrame([(pdf.GRADE.iloc[0], pdf.DIVISION.iloc[0], pdf.VALUE.sum(), )])
+                ...
+                >>> df = session.createDataFrame([('A', 2, 11.0), ('A', 2, 13.9), ('B', 5, 5.0), ('B', 2, 12.1)],
+                ...                              schema=["grade", "division", "value"])
+                >>> df.group_by([df.grade, df.division] ).applyInPandas(
+                ...     group_sum,
+                ...     output_schema=StructType([StructField("grade", StringType()),
+                ...                                        StructField("division", IntegerType()),
+                ...                                        StructField("sum", DoubleType())]),
+                ...                is_permanent=True, stage_location="@mystage", name="group_sum_in_pandas", replace=True
+                ...            ).order_by("sum").show()
+                --------------------------------
+                |"GRADE"  |"DIVISION"  |"SUM"  |
+                --------------------------------
+                |B        |5           |5.0    |
+                |B        |2           |12.1   |
+                |A        |2           |24.9   |
+                --------------------------------
+                <BLANKLINE>
+
+        See Also:
+            - :class:`~snowflake.snowpark.udtf.UDTFRegistration`
+            - :func:`~snowflake.snowpark.functions.pandas_udtf`
+        """
+
+        class _ApplyInPandas:
+            def end_partition(self, pdf: pandas.DataFrame) -> pandas.DataFrame:
+                return func(pdf)
+
+        # for vectorized UDTF
+        _ApplyInPandas.end_partition._sf_vectorized_input = pandas.DataFrame
+
+        # The assumption here is that we send all columns of the dataframe in the apply_in_pandas
+        # function so the inferred input types are the types of each column in the dataframe.
+        kwargs["input_types"] = kwargs.get(
+            "input_types", [field.datatype for field in self._df.schema.fields]
+        )
+
+        kwargs["input_names"] = kwargs.get(
+            "input_names", [field.name for field in self._df.schema.fields]
+        )
+
+        _apply_in_pandas_udtf = self._df._session.udtf.register(
+            _ApplyInPandas,
+            output_schema=output_schema,
+            **kwargs,
+        )
+        partition_by = [functions.col(expr) for expr in self._grouping_exprs]
+
+        return self._df.select(
+            _apply_in_pandas_udtf(*self._df.columns).over(partition_by=partition_by)
+        )
+
+    applyInPandas = apply_in_pandas
+
+    def pivot(
+        self, pivot_col: ColumnOrName, values: Iterable[LiteralType]
+    ) -> "RelationalGroupedDataFrame":
+        """Rotates this DataFrame by turning unique values from one column in the input
+        expression into multiple columns and aggregating results where required on any
+        remaining column values.
+
+        Only one aggregate is supported with pivot.
+
+        Args:
+            pivot_col: The column or name of the column to use.
+            values: A list of values in the column.
+
+        Example::
+
+            >>> create_result = session.sql('''create or replace temp table monthly_sales(empid int, team text, amount int, month text)
+            ... as select * from values
+            ... (1, 'A', 10000, 'JAN'),
+            ... (1, 'B', 400, 'JAN'),
+            ... (2, 'A', 4500, 'JAN'),
+            ... (2, 'A', 35000, 'JAN'),
+            ... (1, 'B', 5000, 'FEB'),
+            ... (1, 'A', 3000, 'FEB'),
+            ... (2, 'B', 200, 'FEB') ''').collect()
+            >>> df = session.table("monthly_sales")
+            >>> df.group_by("empid").pivot("month", ['JAN', 'FEB']).sum("amount").sort(df["empid"]).show()
+            -------------------------------
+            |"EMPID"  |"'JAN'"  |"'FEB'"  |
+            -------------------------------
+            |1        |10400    |8000     |
+            |2        |39500    |200      |
+            -------------------------------
+            <BLANKLINE>
+
+            >>> df.group_by(["empid", "team"]).pivot("month", ['JAN', 'FEB']).sum("amount").sort("empid", "team").show()
+            ----------------------------------------
+            |"EMPID"  |"TEAM"  |"'JAN'"  |"'FEB'"  |
+            ----------------------------------------
+            |1        |A       |10000    |3000     |
+            |1        |B       |400      |5000     |
+            |2        |A       |39500    |NULL     |
+            |2        |B       |NULL     |200      |
+            ----------------------------------------
+            <BLANKLINE>
+        """
+        if not values:
+            raise ValueError("values cannot be empty")
+        pc = self._df._convert_cols_to_exprs(
+            "RelationalGroupedDataFrame.pivot()", pivot_col
+        )
+        value_exprs = [
+            v._expression if isinstance(v, Column) else Literal(v) for v in values
+        ]
+        self._group_type = _PivotType(pc[0], value_exprs)
+        return self
+
     @relational_group_df_api_usage
     def avg(self, *cols: ColumnOrName) -> DataFrame:
         """Return the average for the specified numeric columns."""
