@@ -16,6 +16,8 @@ from snowflake.snowpark._internal.telemetry import TelemetryField
 from snowflake.snowpark._internal.type_utils import convert_sp_to_sf_type
 from snowflake.snowpark._internal.udf_utils import (
     UDFColumn,
+    check_execute_as_arg,
+    check_python_runtime_version,
     check_register_args,
     cleanup_failed_permanent_registration,
     create_python_udf_or_sp,
@@ -26,18 +28,16 @@ from snowflake.snowpark._internal.udf_utils import (
     resolve_imports_and_packages,
 )
 from snowflake.snowpark._internal.utils import TempObjectType
-from snowflake.snowpark.types import DataType
+from snowflake.snowpark.types import DataType, StructType
 
 # Python 3.8 needs to use typing.Iterable because collections.abc.Iterable is not subscriptable
 # Python 3.9 can use both
 # Python 3.10 needs to use collections.abc.Iterable because typing.Iterable is removed
-try:
+if sys.version_info <= (3, 9):
     from typing import Iterable
-except ImportError:
+else:
     from collections.abc import Iterable
 
-EXECUTE_AS_WHITELIST = frozenset(["owner", "caller"])
-
 
 class StoredProcedure:
     """
@@ -72,6 +72,7 @@ class StoredProcedure:
         self._input_types = input_types
         self._execute_as = execute_as
         self._anonymous_sp_sql = anonymous_sp_sql
+        self._is_return_table = isinstance(return_type, StructType)
 
     def __call__(
         self,
@@ -100,9 +101,14 @@ class StoredProcedure:
 
         if self._anonymous_sp_sql:
             call_sql = generate_call_python_sp_sql(session, self.name, *args)
-            return session.sql(f"{self._anonymous_sp_sql}{call_sql}").collect()[0][0]
+            df = session.sql(f"{self._anonymous_sp_sql}{call_sql}")
+            if self._is_return_table:
+                return df
+            return df.collect()[0][0]
         else:
-            return session.call(self.name, *args)
+            return session._call(
+                self.name, *args, is_return_table=self._is_return_table
+            )
 
 
 class StoredProcedureRegistration:
@@ -181,6 +187,10 @@ class StoredProcedureRegistration:
         3. Currently calling stored procedure that requires VARIANT and GEOGRAPHY input types is not supported
         in snowpark API.
 
+        4. Dataframe returned from :meth:`~snowflake.snowpark.Session.call` does not support stacking dataframe
+        operations when sql simplifier is disabled, and output columns in return type for the table stored
+        procedure are not defined.
+
     Example 1
         Use stored procedure to copy data from one table to another::
 
@@ -337,6 +347,54 @@ class StoredProcedureRegistration:
             >>> mod5_sp(2)
             2
 
+    Example 9
+        Creating a table stored procedure with return type while defining return columns and datatypes::
+
+            >>> from snowflake.snowpark.types import IntegerType, StructField, StructType
+            >>> @sproc(return_type=StructType([StructField("A", IntegerType()), StructField("B", IntegerType())]), input_types=[IntegerType(), IntegerType()])
+            ... def select_sp(session_, x, y):
+            ...     return session_.sql(f"SELECT {x} as A, {y} as B")
+            ...
+            >>> select_sp(1, 2).show()
+            -------------
+            |"A"  |"B"  |
+            -------------
+            |1    |2    |
+            -------------
+            <BLANKLINE>
+
+    Example 10
+        Creating a table stored procedure with return type with free return columns::
+
+            >>> from snowflake.snowpark.types import IntegerType, StructType
+            >>> @sproc(return_type=StructType(), input_types=[IntegerType(), IntegerType()])
+            ... def select_sp(session_, x, y):
+            ...     return session_.sql(f"SELECT {x} as A, {y} as B")
+            ...
+            >>> select_sp(1, 2).show()
+            -------------
+            |"A"  |"B"  |
+            -------------
+            |1    |2    |
+            -------------
+            <BLANKLINE>
+
+    Example 9
+        Creating a table stored procedure using implicit type hints::
+
+            >>> from snowflake.snowpark.dataframe import DataFrame
+            >>> @sproc
+            ... def select_sp(session_: snowflake.snowpark.Session, x: int, y: int) -> DataFrame:
+            ...     return session_.sql(f"SELECT {x} as A, {y} as B")
+            ...
+            >>> select_sp(1, 2).show()
+            -------------
+            |"A"  |"B"  |
+            -------------
+            |1    |2    |
+            -------------
+            <BLANKLINE>
+
     See Also:
         - :class:`snowflake.snowpark.udf.UDFRegistration`
         - :func:`~snowflake.snowpark.functions.sproc`
@@ -379,6 +437,8 @@ class StoredProcedureRegistration:
         parallel: int = 4,
         execute_as: typing.Literal["caller", "owner"] = "owner",
         strict: bool = False,
+        external_access_integrations: Optional[List[str]] = None,
+        secrets: Optional[Dict[str, str]] = None,
         *,
         statement_params: Optional[Dict[str, str]] = None,
         source_code_display: bool = True,
@@ -445,6 +505,13 @@ class StoredProcedureRegistration:
                 The source code is dynamically generated therefore it may not be identical to how the
                 `func` is originally defined. The default is ``True``.
                 If it is ``False``, source code will not be generated or displayed.
+            external_access_integrations: The names of one or more external access integrations. Each
+                integration you specify allows access to the external network locations and secrets
+                the integration specifies.
+            secrets: The key-value pairs of string types of secrets used to authenticate the external network location.
+                The secrets can be accessed from handler code. The secrets specified as values must
+                also be specified in the external access integration and the keys are strings used to
+                retrieve the secrets using secret API.
 
         See Also:
             - :func:`~snowflake.snowpark.functions.sproc`
@@ -455,15 +522,8 @@ class StoredProcedureRegistration:
                 "Invalid function: not a function or callable "
                 f"(__call__ is not defined): {type(func)}"
             )
-        if (
-            not isinstance(execute_as, str)
-            or execute_as.lower() not in EXECUTE_AS_WHITELIST
-        ):
-            raise TypeError(
-                f"'execute_as' value '{execute_as}' is invalid, choose from "
-                f"{', '.join(EXECUTE_AS_WHITELIST, )}"
-            )
 
+        check_execute_as_arg(execute_as)
         check_register_args(
             TempObjectType.PROCEDURE, name, is_permanent, stage_location, parallel
         )
@@ -481,11 +541,18 @@ class StoredProcedureRegistration:
             if_not_exists,
             parallel,
             strict,
+            external_access_integrations=external_access_integrations,
+            secrets=secrets,
             statement_params=statement_params,
             execute_as=execute_as,
             api_call_source="StoredProcedureRegistration.register",
             source_code_display=source_code_display,
             anonymous=kwargs.get("anonymous", False),
+            is_permanent=is_permanent,
+            # force_inline_code avoids uploading python file
+            # when we know the code is not too large. This is useful
+            # in Pandas API to create stored procedures not registered by users.
+            force_inline_code=kwargs.get("force_inline_code", False),
         )
 
     def register_from_file(
@@ -502,10 +569,14 @@ class StoredProcedureRegistration:
         replace: bool = False,
         if_not_exists: bool = False,
         parallel: int = 4,
+        execute_as: typing.Literal["caller", "owner"] = "owner",
         strict: bool = False,
+        external_access_integrations: Optional[List[str]] = None,
+        secrets: Optional[Dict[str, str]] = None,
         *,
         statement_params: Optional[Dict[str, str]] = None,
         source_code_display: bool = True,
+        skip_upload_on_content_match: bool = False,
     ) -> StoredProcedure:
         """
         Registers a Python function as a Snowflake Python stored procedure from a Python or zip file,
@@ -564,6 +635,8 @@ class StoredProcedureRegistration:
                 command. The default value is 4 and supported values are from 1 to 99.
                 Increasing the number of threads can improve performance when uploading
                 large stored procedure files.
+            execute_as: What permissions should the procedure have while executing. This
+                supports caller, or owner for now.
             strict: Whether the created stored procedure is strict. A strict stored procedure will not invoke
                 the stored procedure if any input is null. Instead, a null value will always be returned. Note
                 that the stored procedure might still return null for non-null inputs.
@@ -572,6 +645,16 @@ class StoredProcedureRegistration:
                 The source code is dynamically generated therefore it may not be identical to how the
                 `func` is originally defined. The default is ``True``.
                 If it is ``False``, source code will not be generated or displayed.
+            skip_upload_on_content_match: When set to ``True`` and a version of source file already exists on stage, the given source
+                file will be uploaded to stage only if the contents of the current file differ from the remote file on stage. Defaults
+                to ``False``.
+            external_access_integrations: The names of one or more external access integrations. Each
+                integration you specify allows access to the external network locations and secrets
+                the integration specifies.
+            secrets: The key-value pairs of string types of secrets used to authenticate the external network location.
+                The secrets can be accessed from handler code. The secrets specified as values must
+                also be specified in the external access integration and the keys are strings used to
+                retrieve the secrets using secret API.
 
         Note::
             The type hints can still be extracted from the source Python file if they
@@ -587,6 +670,7 @@ class StoredProcedureRegistration:
         check_register_args(
             TempObjectType.PROCEDURE, name, is_permanent, stage_location, parallel
         )
+        check_execute_as_arg(execute_as)
 
         # register stored procedure
         return self._do_register_sp(
@@ -601,9 +685,14 @@ class StoredProcedureRegistration:
             if_not_exists,
             parallel,
             strict,
+            external_access_integrations=external_access_integrations,
+            secrets=secrets,
             statement_params=statement_params,
+            execute_as=execute_as,
             api_call_source="StoredProcedureRegistration.register_from_file",
             source_code_display=source_code_display,
+            skip_upload_on_content_match=skip_upload_on_content_match,
+            is_permanent=is_permanent,
         )
 
     def _do_register_sp(
@@ -625,6 +714,11 @@ class StoredProcedureRegistration:
         execute_as: typing.Literal["caller", "owner"] = "owner",
         anonymous: bool = False,
         api_call_source: str,
+        skip_upload_on_content_match: bool = False,
+        is_permanent: bool = False,
+        external_access_integrations: Optional[List[str]] = None,
+        secrets: Optional[Dict[str, str]] = None,
+        force_inline_code: bool = False,
     ) -> StoredProcedure:
         (
             udf_name,
@@ -656,6 +750,7 @@ class StoredProcedureRegistration:
             all_imports,
             all_packages,
             upload_file_stage_location,
+            custom_python_runtime_version_allowed,
         ) = resolve_imports_and_packages(
             self._session,
             TempObjectType.PROCEDURE,
@@ -668,8 +763,16 @@ class StoredProcedureRegistration:
             parallel,
             statement_params=statement_params,
             source_code_display=source_code_display,
+            skip_upload_on_content_match=skip_upload_on_content_match,
+            is_permanent=is_permanent,
+            force_inline_code=force_inline_code,
         )
 
+        if not custom_python_runtime_version_allowed:
+            check_python_runtime_version(
+                self._session._runtime_version_from_requirement
+            )
+
         anonymous_sp_sql = None
         if anonymous:
             anonymous_sp_sql = generate_anonymous_python_sp_sql(
@@ -681,6 +784,9 @@ class StoredProcedureRegistration:
                 all_packages=all_packages,
                 inline_python_code=code,
                 strict=strict,
+                runtime_version=self._session._runtime_version_from_requirement,
+                external_access_integrations=external_access_integrations,
+                secrets=secrets,
             )
         else:
             raised = False
@@ -694,13 +800,15 @@ class StoredProcedureRegistration:
                     object_name=udf_name,
                     all_imports=all_imports,
                     all_packages=all_packages,
-                    is_temporary=stage_location is None,
+                    is_permanent=is_permanent,
                     replace=replace,
                     if_not_exists=if_not_exists,
                     inline_python_code=code,
                     execute_as=execute_as,
                     api_call_source=api_call_source,
                     strict=strict,
+                    external_access_integrations=external_access_integrations,
+                    secrets=secrets,
                 )
             # an exception might happen during registering a stored procedure
             # (e.g., a dependency might not be found on the stage),
