@@ -3,14 +3,12 @@
 # Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
 #
 
+import sys
+from logging import getLogger
 from typing import Dict, List, NamedTuple, Optional, Union, overload
 
 import snowflake.snowpark
 from snowflake.snowpark._internal.analyzer.binary_plan_node import create_join_type
-from snowflake.snowpark._internal.analyzer.select_statement import (
-    SelectableEntity,
-    SelectStatement,
-)
 from snowflake.snowpark._internal.analyzer.snowflake_plan_node import UnresolvedRelation
 from snowflake.snowpark._internal.analyzer.table_merge_expression import (
     DeleteMergeExpression,
@@ -20,6 +18,7 @@ from snowflake.snowpark._internal.analyzer.table_merge_expression import (
     TableUpdate,
     UpdateMergeExpression,
 )
+from snowflake.snowpark._internal.analyzer.unary_plan_node import Sample
 from snowflake.snowpark._internal.error_message import SnowparkClientExceptionMessages
 from snowflake.snowpark._internal.telemetry import add_api_call, set_api_call_source
 from snowflake.snowpark._internal.type_utils import ColumnOrLiteral
@@ -30,11 +29,13 @@ from snowflake.snowpark.row import Row
 # Python 3.8 needs to use typing.Iterable because collections.abc.Iterable is not subscriptable
 # Python 3.9 can use both
 # Python 3.10 needs to use collections.abc.Iterable because typing.Iterable is removed
-try:
+if sys.version_info <= (3, 9):
     from typing import Iterable
-except ImportError:
+else:
     from collections.abc import Iterable
 
+_logger = getLogger(__name__)
+
 
 class UpdateResult(NamedTuple):
     """Result of updating rows in a :class:`Table`."""
@@ -98,8 +99,8 @@ class WhenMatchedClause:
             >>> source = session.create_dataframe([(10, "new")], schema=["key", "value"])
             >>> target.merge(source, (target["key"] == source["key"]) & (target["value"] == lit("too_old")), [when_matched().update({"value": source["value"]})])
             MergeResult(rows_inserted=0, rows_updated=1, rows_deleted=0)
-            >>> target.collect() # the value in the table is updated
-            [Row(KEY=10, VALUE='old'), Row(KEY=10, VALUE='new'), Row(KEY=11, VALUE='old')]
+            >>> target.sort("key", "value").collect() # the value in the table is updated
+            [Row(KEY=10, VALUE='new'), Row(KEY=10, VALUE='old'), Row(KEY=11, VALUE='old')]
 
         Note:
             An exception will be raised if this method or :meth:`WhenMatchedClause.delete`
@@ -191,22 +192,24 @@ class WhenNotMatchedClause:
             >>> # For all such rows, insert a row into target whose ley and value
             >>> # are assigned to the key and value of the not matched row.
             >>> from snowflake.snowpark.functions import when_not_matched
-            >>> target_df = session.create_dataframe([(10, "old"), (10, "too_old"), (11, "old")], schema=["key", "value"])
+            >>> from snowflake.snowpark.types import IntegerType, StringType, StructField, StructType
+            >>> schema = StructType([StructField("key", IntegerType()), StructField("value", StringType())])
+            >>> target_df = session.create_dataframe([(10, "old"), (10, "too_old"), (11, "old")], schema=schema)
             >>> target_df.write.save_as_table("my_table", mode="overwrite", table_type="temporary")
             >>> target = session.table("my_table")
-            >>> source = session.create_dataframe([(12, "new")], schema=["key", "value"])
+            >>> source = session.create_dataframe([(12, "new")], schema=schema)
             >>> target.merge(source, target["key"] == source["key"], [when_not_matched().insert([source["key"], source["value"]])])
             MergeResult(rows_inserted=1, rows_updated=0, rows_deleted=0)
-            >>> target.collect() # the rows are inserted
-            [Row(KEY=12, VALUE='new'), Row(KEY=10, VALUE='old'), Row(KEY=10, VALUE='too_old'), Row(KEY=11, VALUE='old')]
+            >>> target.sort("key", "value").collect() # the rows are inserted
+            [Row(KEY=10, VALUE='old'), Row(KEY=10, VALUE='too_old'), Row(KEY=11, VALUE='old'), Row(KEY=12, VALUE='new')]
 
             >>> # For all such rows, insert a row into target whose key is
             >>> # assigned to the key of the not matched row.
             >>> target_df.write.save_as_table("my_table", mode="overwrite", table_type="temporary")
             >>> target.merge(source, target["key"] == source["key"], [when_not_matched().insert({"key": source["key"]})])
             MergeResult(rows_inserted=1, rows_updated=0, rows_deleted=0)
-            >>> target.collect() # the rows are inserted
-            [Row(KEY=12, VALUE=None), Row(KEY=10, VALUE='old'), Row(KEY=10, VALUE='too_old'), Row(KEY=11, VALUE='old')]
+            >>> target.sort("key", "value").collect() # the rows are inserted
+            [Row(KEY=10, VALUE='old'), Row(KEY=10, VALUE='too_old'), Row(KEY=11, VALUE='old'), Row(KEY=12, VALUE=None)]
 
         Note:
             An exception will be raised if this method is called more than once
@@ -271,8 +274,10 @@ class Table(DataFrame):
         self.table_name: str = table_name  #: The table name
 
         if self._session.sql_simplifier_enabled:
-            self._select_statement = SelectStatement(
-                from_=SelectableEntity(table_name, analyzer=session._analyzer),
+            self._select_statement = session._analyzer.create_select_statement(
+                from_=session._analyzer.create_selectable_entity(
+                    table_name, analyzer=session._analyzer
+                ),
                 analyzer=session._analyzer,
             )
         # By default, the set the initial API call to say 'Table.__init__' since
@@ -294,7 +299,7 @@ class Table(DataFrame):
         frac: Optional[float] = None,
         n: Optional[int] = None,
         *,
-        seed: Optional[float] = None,
+        seed: Optional[int] = None,
         sampling_method: Optional[str] = None,
     ) -> "DataFrame":
         """Samples rows based on either the number of rows to be returned or a percentage of rows to be returned.
@@ -332,6 +337,26 @@ class Table(DataFrame):
                 f"'sampling_method' value {sampling_method} must be None or one of 'BERNOULLI', 'ROW', 'SYSTEM', or 'BLOCK'."
             )
 
+        from snowflake.snowpark.mock._connection import MockServerConnection
+
+        if isinstance(self._session._conn, MockServerConnection):
+            if sampling_method in ("SYSTEM", "BLOCK"):
+                _logger.warning(
+                    "[Local Testing] SYSTEM/BLOCK sampling is not supported for Local Testing, falling back to ROW sampling"
+                )
+
+            sample_plan = Sample(
+                self._plan, probability_fraction=frac, row_count=n, seed=seed
+            )
+            return self._with_plan(
+                self._session._analyzer.create_select_statement(
+                    from_=self._session._analyzer.create_select_snowflake_plan(
+                        sample_plan, analyzer=self._session._analyzer
+                    ),
+                    analyzer=self._session._analyzer,
+                )
+            )
+
         # The analyzer will generate a sql with subquery. So we build the sql directly without using the analyzer.
         sampling_method_text = sampling_method or ""
         frac_or_rowcount_text = str(frac * 100.0) if frac is not None else f"{n} ROWS"
@@ -400,14 +425,14 @@ class Table(DataFrame):
             >>> # to the summation of column "a" and column "b"
             >>> t.update({"b": 0, "a": t.a + t.b})
             UpdateResult(rows_updated=6, multi_joined_rows_updated=0)
-            >>> t.collect()
+            >>> t.sort("a", "b").collect()
             [Row(A=2, B=0), Row(A=3, B=0), Row(A=3, B=0), Row(A=4, B=0), Row(A=4, B=0), Row(A=5, B=0)]
 
             >>> # update all rows in column "b" to 0 where column "a" has value 1
             >>> target_df.write.save_as_table("my_table", mode="overwrite", table_type="temporary")
             >>> t.update({"b": 0}, t["a"] == 1)
             UpdateResult(rows_updated=2, multi_joined_rows_updated=0)
-            >>> t.collect()
+            >>> t.sort("a", "b").collect()
             [Row(A=1, B=0), Row(A=1, B=0), Row(A=2, B=1), Row(A=2, B=2), Row(A=3, B=1), Row(A=3, B=2)]
 
             >>> # update all rows in column "b" to 0 where column "a" in this
@@ -416,7 +441,7 @@ class Table(DataFrame):
             >>> source_df = session.create_dataframe([1, 2, 3, 4], schema=["a"])
             >>> t.update({"b": 0}, t["a"] == source_df.a, source_df)
             UpdateResult(rows_updated=6, multi_joined_rows_updated=0)
-            >>> t.collect()
+            >>> t.sort("a", "b").collect()
             [Row(A=1, B=0), Row(A=1, B=0), Row(A=2, B=0), Row(A=2, B=0), Row(A=3, B=0), Row(A=3, B=0)]
         """
         if source:
@@ -505,7 +530,7 @@ class Table(DataFrame):
             >>> target_df.write.save_as_table("my_table", mode="overwrite", table_type="temporary")
             >>> t.delete(t["a"] == 1)
             DeleteResult(rows_deleted=2)
-            >>> t.collect()
+            >>> t.sort("a", "b").collect()
             [Row(A=2, B=1), Row(A=2, B=2), Row(A=3, B=1), Row(A=3, B=2)]
 
             >>> # delete all rows in this table where column "a" in this
@@ -514,7 +539,7 @@ class Table(DataFrame):
             >>> source_df = session.create_dataframe([2, 3, 4, 5], schema=["a"])
             >>> t.delete(t["a"] == source_df.a, source_df)
             DeleteResult(rows_deleted=4)
-            >>> t.collect()
+            >>> t.sort("a", "b").collect()
             [Row(A=1, B=1), Row(A=1, B=2)]
         """
         if source:
@@ -598,15 +623,17 @@ class Table(DataFrame):
         Example::
 
             >>> from snowflake.snowpark.functions import when_matched, when_not_matched
-            >>> target_df = session.create_dataframe([(10, "old"), (10, "too_old"), (11, "old")], schema=["key", "value"])
+            >>> from snowflake.snowpark.types import IntegerType, StringType, StructField, StructType
+            >>> schema = StructType([StructField("key", IntegerType()), StructField("value", StringType())])
+            >>> target_df = session.create_dataframe([(10, "old"), (10, "too_old"), (11, "old")], schema=schema)
             >>> target_df.write.save_as_table("my_table", mode="overwrite", table_type="temporary")
             >>> target = session.table("my_table")
-            >>> source = session.create_dataframe([(10, "new"), (12, "new"), (13, "old")], schema=["key", "value"])
+            >>> source = session.create_dataframe([(10, "new"), (12, "new"), (13, "old")], schema=schema)
             >>> target.merge(source, (target["key"] == source["key"]) & (target["value"] == "too_old"),
             ...              [when_matched().update({"value": source["value"]}), when_not_matched().insert({"key": source["key"]})])
             MergeResult(rows_inserted=2, rows_updated=1, rows_deleted=0)
-            >>> target.collect()
-            [Row(KEY=12, VALUE=None), Row(KEY=13, VALUE=None), Row(KEY=10, VALUE='old'), Row(KEY=10, VALUE='new'), Row(KEY=11, VALUE='old')]
+            >>> target.sort("key", "value").collect()
+            [Row(KEY=10, VALUE='new'), Row(KEY=10, VALUE='old'), Row(KEY=11, VALUE='old'), Row(KEY=12, VALUE=None), Row(KEY=13, VALUE=None)]
         """
         inserted, updated, deleted = False, False, False
         merge_exprs = []
@@ -659,6 +686,12 @@ class Table(DataFrame):
         Note that subsequent operations such as :meth:`DataFrame.select`, :meth:`DataFrame.collect` on this ``Table`` instance and the derived DataFrame will raise errors because the underlying
         table in the Snowflake database no longer exists.
         """
-        self._session.sql(
-            f"drop table {self.table_name}"
-        )._internal_collect_with_tag_no_telemetry()
+        from snowflake.snowpark.mock._connection import MockServerConnection
+
+        if isinstance(self._session._conn, MockServerConnection):
+            # only mock connection has entity_registry
+            self._session._conn.entity_registry.drop_table(self.table_name)
+        else:
+            self._session.sql(
+                f"drop table {self.table_name}"
+            )._internal_collect_with_tag_no_telemetry()
