@@ -5,7 +5,6 @@
 """Contains table function related classes."""
 from typing import Dict, Iterable, List, Optional, Tuple, Union
 
-from snowflake.snowpark._internal.analyzer.analyzer_utils import quote_name
 from snowflake.snowpark._internal.analyzer.sort_expression import Ascending, SortOrder
 from snowflake.snowpark._internal.analyzer.table_function import (
     NamedArgumentsTableFunction,
@@ -14,8 +13,9 @@ from snowflake.snowpark._internal.analyzer.table_function import (
     TableFunctionPartitionSpecDefinition,
 )
 from snowflake.snowpark._internal.type_utils import ColumnOrName
-from snowflake.snowpark._internal.utils import validate_object_name
+from snowflake.snowpark._internal.utils import quote_name, validate_object_name
 from snowflake.snowpark.column import Column, _to_col_if_str
+from snowflake.snowpark.types import ArrayType, MapType
 
 from ._internal.analyzer.snowflake_plan import SnowflakePlan
 
@@ -39,6 +39,7 @@ class TableFunctionCall:
         if func_arguments and func_named_arguments:
             raise ValueError("A table function shouldn't have both args and named args")
         self.name: str = func_name  #: The table function name
+        self.user_visible_name: str = func_name
         self.arguments: Iterable[
             ColumnOrName
         ] = func_arguments  #: The positional arguments used to call this table function.
@@ -131,6 +132,18 @@ class TableFunctionCall:
     as_ = alias
 
 
+class _ExplodeFunctionCall(TableFunctionCall):
+    """Internal class to identify explode function call as a special instance of TableFunctionCall"""
+
+    def __init__(self, col: ColumnOrName, outer: Column) -> None:
+        super().__init__("flatten", input=col, outer=outer)
+        if isinstance(col, Column):
+            self.col = col._expression.name
+        else:
+            self.col = quote_name(col)
+        self.user_visible_name: str = "explode"
+
+
 def _create_order_by_expression(e: Union[str, Column]) -> SortOrder:
     if isinstance(e, str):
         return SortOrder(Column(e)._expression, Ascending())
@@ -207,7 +220,7 @@ def _get_cols_after_join_table(
     func_expr: TableFunctionExpression,
     current_plan: SnowflakePlan,
     join_plan: SnowflakePlan,
-) -> Tuple[List, List]:
+) -> Tuple[List, List, List]:
     def get_column_names_from_plan(plan: SnowflakePlan) -> List[str]:
         return [attr.name for attr in plan.output]
 
@@ -216,20 +229,59 @@ def _get_cols_after_join_table(
     cols_after_join = get_column_names_from_plan(join_plan)
     aliases = func_expr.aliases
 
-    new_cols = [col for col in cols_after_join if col not in cols_before_join]
+    new_cols_names = cols_after_join[len(cols_before_join) :]
     old_cols = [Column(col)._named() for col in cols_before_join]
+    alias_cols = [Column(col)._named() for col in aliases] if aliases else []
 
     if aliases:
-        if len(new_cols) != len(aliases):
+        if len(new_cols_names) != len(aliases):
             raise ValueError(
                 f"The number of aliases should be same as the number of cols added by table function. "
-                f"Columns added by table function are {new_cols} and aliases given are {aliases}"
+                f"Columns added by table function are {new_cols_names} and aliases given are {aliases}"
             )
         new_cols = [
-            Column(col).alias(alias_col)._named()
-            for col, alias_col in zip(new_cols, aliases)
+            Column(name).alias(alias_name)._named()
+            for name, alias_name in zip(new_cols_names, aliases)
         ]
     else:
-        new_cols = [Column(col)._named() for col in new_cols]
+        new_cols = [Column(col)._named() for col in new_cols_names]
 
-    return old_cols, new_cols
+    return old_cols, new_cols, alias_cols
+
+
+def _get_cols_after_explode_join(
+    func: _ExplodeFunctionCall, plan: SnowflakePlan
+) -> Tuple[List, List]:
+    explode_col_type = plan.output_dict.get(func.col, [None])[0]
+
+    cols = []
+    aliases = func._aliases
+    alias_cols = [Column(col)._named() for col in aliases] if aliases else []
+    if isinstance(explode_col_type, ArrayType):
+        if aliases:
+            if len(aliases) != 1:
+                raise ValueError(
+                    f"Invalid number of aliases given for explode. Expecting 1, got {len(aliases)}"
+                )
+            cols.append(Column("VALUE").alias(aliases[0])._named())
+        else:
+            cols.append(Column("VALUE")._named())
+    elif isinstance(explode_col_type, MapType):
+        if aliases:
+            if len(aliases) != 2:
+                raise ValueError(
+                    f"Invalid number of aliases given for explode. Expecting 2, got {len(aliases)}"
+                )
+            cols.extend(
+                [
+                    Column("KEY").as_(aliases[0])._named(),
+                    Column("VALUE").as_(aliases[1])._named(),
+                ]
+            )
+        else:
+            cols.extend([Column("KEY")._named(), Column("VALUE")._named()])
+    else:
+        raise ValueError(
+            f"Invalid column type for explode({func.col}). Expected ArrayType() or MapType(), got {explode_col_type}"
+        )
+    return cols, alias_cols
